"""
This module implements :class:`AnalogSignal`, an array of analog signals.

:class:`AnalogSignal` inherits from :class:`basesignal.BaseSignal` which
derives from :class:`BaseNeo`, and from :class:`quantities.Quantity`which
in turn inherits from :class:`numpy.array`.

Inheritance from :class:`numpy.array` is explained here:
http://docs.scipy.org/doc/numpy/user/basics.subclassing.html

In brief:
* Initialization of a new object from constructor happens in :meth:`__new__`.
This is where user-specified attributes are set.

* :meth:`__array_finalize__` is called for all new objects, including those
created by slicing. This is where attributes are copied over from
the old object.
"""

import logging
from copy import deepcopy

import numpy as np
import quantities as pq

from neo.core.baseneo import BaseNeo, MergeError, merge_annotations, intersect_annotations
from neo.core.dataobject import DataObject
from neo.core.basesignal import BaseSignal

logger = logging.getLogger("Neo")


def _get_sampling_rate(sampling_rate, sampling_period):
    """
    Gets the sampling_rate from either the sampling_period or the
    sampling_rate, or makes sure they match if both are specified
    """
    if sampling_period is None:
        if sampling_rate is None:
            raise ValueError("You must provide either the sampling rate or " + "sampling period")
    elif sampling_rate is None:
        sampling_rate = 1.0 / sampling_period
    elif sampling_period != 1.0 / sampling_rate:
        raise ValueError("The sampling_rate has to be 1/sampling_period")
    if not hasattr(sampling_rate, "units"):
        raise TypeError("Sampling rate/sampling period must have units")
    return sampling_rate


def _new_AnalogSignalArray(
    cls,
    signal,
    units=None,
    dtype=None,
    copy=True,
    t_start=0 * pq.s,
    sampling_rate=None,
    sampling_period=None,
    name=None,
    file_origin=None,
    description=None,
    array_annotations=None,
    annotations=None,
    segment=None,
):
    """
    A function to map AnalogSignal.__new__ to function that
        does not do the unit checking. This is needed for pickle to work.
    """
    obj = cls(
        signal=signal,
        units=units,
        dtype=dtype,
        copy=copy,
        t_start=t_start,
        sampling_rate=sampling_rate,
        sampling_period=sampling_period,
        name=name,
        file_origin=file_origin,
        description=description,
        array_annotations=array_annotations,
        **annotations,
    )
    obj.segment = segment
    return obj


class AnalogSignal(BaseSignal):
    """
    Array of one or more continuous analog signals.

    A representation of several continuous, analog signals that
    have the same duration, sampling rate and start time.
    Basically, it is a 2D array: dim 0 is time, dim 1 is
    channel index

    Inherits from :class:`quantities.Quantity`, which in turn inherits from
    :class:`numpy.ndarray`.

    *Usage*::

        >>> from neo.core import AnalogSignal
        >>> import quantities as pq
        >>>
        >>> sigarr = AnalogSignal([[1, 2, 3], [4, 5, 6]], units='V',
        ...                            sampling_rate=1*pq.Hz)
        >>>
        >>> sigarr
        <AnalogSignal(array([[1, 2, 3],
              [4, 5, 6]]) * mV, [0.0 s, 2.0 s], sampling rate: 1.0 Hz)>
        >>> sigarr[:,1]
        <AnalogSignal(array([2, 5]) * V, [0.0 s, 2.0 s],
            sampling rate: 1.0 Hz)>
        >>> sigarr[1, 1]
        array(5) * V

    *Required attributes/properties*:
        :signal: (quantity array 2D, numpy array 2D, or list (data, channel))
            The data itself.
        :units: (quantity units) Required if the signal is a list or NumPy
                array, not if it is a :class:`Quantity`
        :t_start: (quantity scalar) Time when signal begins
        :sampling_rate: *or* **sampling_period** (quantity scalar) Number of
                                               samples per unit time or
                                               interval between two samples.
                                               If both are specified, they are
                                               checked for consistency.

    *Recommended attributes/properties*:
        :name: (str) A label for the dataset.
        :description: (str) Text description.
        :file_origin: (str) Filesystem path or URL of the original data file.

    *Optional attributes/properties*:
        :dtype: (numpy dtype or str) Override the dtype of the signal array.
        :copy: (bool) True by default.
        :array_annotations: (dict) Dict mapping strings to numpy arrays containing annotations \
                                   for all data points

    Note: Any other additional arguments are assumed to be user-specific
    metadata and stored in :attr:`annotations`.

    *Properties available on this object*:
        :sampling_rate: (quantity scalar) Number of samples per unit time.
            (1/:attr:`sampling_period`)
        :sampling_period: (quantity scalar) Interval between two samples.
            (1/:attr:`quantity scalar`)
        :duration: (Quantity) Signal duration, read-only.
            (size * :attr:`sampling_period`)
        :t_stop: (quantity scalar) Time when signal ends, read-only.
            (:attr:`t_start` + :attr:`duration`)
        :times: (quantity 1D) The time points of each sample of the signal,
            read-only.
            (:attr:`t_start` + arange(:attr:`shape`[0])/:attr:`sampling_rate`)

    *Slicing*:
        :class:`AnalogSignal` objects can be sliced. When taking a single
        column (dimension 0, e.g. [0, :]) or a single element,
        a :class:`~quantities.Quantity` is returned.
        Otherwise an :class:`AnalogSignal` (actually a view) is
        returned, with the same metadata, except that :attr:`t_start`
        is changed if the start index along dimension 1 is greater than 1.
        Note that slicing an :class:`AnalogSignal` may give a different
        result to slicing the underlying NumPy array since signals
        are always two-dimensional.

    *Operations available on this object*:
        == != + * /

    """

    _parent_objects = ("Segment",)
    _parent_attrs = ("segment",)
    _quantity_attr = "signal"
    _necessary_attrs = (("signal", pq.Quantity, 2), ("sampling_rate", pq.Quantity, 0), ("t_start", pq.Quantity, 0))
    _recommended_attrs = BaseNeo._recommended_attrs

    def __new__(
        cls,
        signal,
        units=None,
        dtype=None,
        copy=True,
        t_start=0 * pq.s,
        sampling_rate=None,
        sampling_period=None,
        name=None,
        file_origin=None,
        description=None,
        array_annotations=None,
        **annotations,
    ):
        """
        Constructs new :class:`AnalogSignal` from data.

        This is called whenever a new class:`AnalogSignal` is created from
        the constructor, but not when slicing.

        __array_finalize__ is called on the new object.
        """
        signal = cls._rescale(signal, units=units)
        obj = pq.Quantity(signal, units=units, dtype=dtype, copy=copy).view(cls)

        if obj.ndim == 1:
            obj.shape = (-1, 1)

        if t_start is None:
            raise ValueError("t_start cannot be None")
        obj._t_start = t_start

        obj._sampling_rate = _get_sampling_rate(sampling_rate, sampling_period)

        obj.segment = None
        return obj

    def __init__(
        self,
        signal,
        units=None,
        dtype=None,
        copy=True,
        t_start=0 * pq.s,
        sampling_rate=None,
        sampling_period=None,
        name=None,
        file_origin=None,
        description=None,
        array_annotations=None,
        **annotations,
    ):
        """
        Initializes a newly constructed :class:`AnalogSignal` instance.
        """
        # This method is only called when constructing a new AnalogSignal,
        # not when slicing or viewing. We use the same call signature
        # as __new__ for documentation purposes. Anything not in the call
        # signature is stored in annotations.

        # Calls parent __init__, which grabs universally recommended
        # attributes and sets up self.annotations
        DataObject.__init__(
            self,
            name=name,
            file_origin=file_origin,
            description=description,
            array_annotations=array_annotations,
            **annotations,
        )

    def __reduce__(self):
        """
        Map the __new__ function onto _new_AnalogSignalArray, so that pickle
        works
        """
        return _new_AnalogSignalArray, (
            self.__class__,
            np.array(self),
            self.units,
            self.dtype,
            True,
            self.t_start,
            self.sampling_rate,
            self.sampling_period,
            self.name,
            self.file_origin,
            self.description,
            self.array_annotations,
            self.annotations,
            self.segment,
        )

    def _array_finalize_spec(self, obj):
        """
        Set default values for attributes specific to :class:`AnalogSignal`.

        Common attributes are defined in
        :meth:`__array_finalize__` in :class:`basesignal.BaseSignal`),
        which is called every time a new signal is created
        and calls this method.
        """
        self._t_start = getattr(obj, "_t_start", 0 * pq.s)
        self._sampling_rate = getattr(obj, "_sampling_rate", None)
        return obj

    def __repr__(self):
        """
        Returns a string representing the :class:`AnalogSignal`.
        """
        return f"<{self.__class__.__name__}({super().__repr__()}, [{self.t_start}, {self.t_stop}], sampling rate: {self.sampling_rate})>"

    def __getitem__(self, i):
        """
        Get the item or slice :attr:`i`.
        """
        if isinstance(i, (int, np.integer)):  # a single point in time across all channels
            obj = super().__getitem__(i)
            obj = pq.Quantity(obj.magnitude, units=obj.units)
        elif isinstance(i, tuple):
            obj = super().__getitem__(i)
            j, k = i
            if isinstance(j, (int, np.integer)):  # extract a quantity array
                obj = pq.Quantity(obj.magnitude, units=obj.units)
            else:
                if isinstance(j, slice):
                    if j.start:
                        obj.t_start = self.t_start + j.start * self.sampling_period
                    if j.step:
                        obj.sampling_period *= j.step
                elif isinstance(j, np.ndarray):
                    raise NotImplementedError("Arrays not yet supported")  # in the general case, would need to return
                    #  IrregularlySampledSignal(Array)
                else:
                    raise TypeError(f"{type(j)} not supported")
                if isinstance(k, (int, np.integer)):
                    obj = obj.reshape(-1, 1)
                elif k is None:
                    # matplotlib _check_1d() calls__getitem__ with (:, None) and
                    # reacts appropriately if an IndexError or ValueError is raised
                    raise IndexError("Cannot add dimensions to an AnalogSignal")
                obj.array_annotate(**deepcopy(self.array_annotations_at_index(k)))
        elif isinstance(i, slice):
            obj = super().__getitem__(i)
            if i.start:
                obj.t_start = self.t_start + i.start * self.sampling_period
            obj.array_annotations = deepcopy(self.array_annotations)
        elif isinstance(i, np.ndarray):
            # Indexing of an AnalogSignal is only consistent if the resulting number of
            # samples is the same for each trace. The time axis for these samples is not
            # guaranteed to be continuous, so returning a Quantity instead of an AnalogSignal here.
            new_time_dims = np.sum(i, axis=0)
            if len(new_time_dims) and all(new_time_dims == new_time_dims[0]):
                obj = np.asarray(self).T.__getitem__(i.T)
                obj = obj.T.reshape(self.shape[1], -1).T
                obj = pq.Quantity(obj, units=self.units)
            else:
                raise IndexError(
                    "indexing of an AnalogSignals needs to keep the same number of " "sample for each trace contained"
                )
        else:
            raise IndexError("index should be an integer, tuple, slice or boolean numpy array")
        return obj

    def __setitem__(self, i, value):
        """
        Set an item or slice defined by :attr:`i` to `value`.
        """
        # because AnalogSignals are always at least two-dimensional,
        # we need to handle the case where `i` is an integer
        if isinstance(i, int):
            i = slice(i, i + 1)
        elif isinstance(i, tuple):
            j, k = i
            if isinstance(k, int):
                i = (j, slice(k, k + 1))
        return super().__setitem__(i, value)

    # sampling_rate attribute is handled as a property so type checking can
    # be done
    @property
    def sampling_rate(self):
        """
        Number of samples per unit time.

        (1/:attr:`sampling_period`)
        """
        return self._sampling_rate

    @sampling_rate.setter
    def sampling_rate(self, rate):
        """
        Setter for :attr:`sampling_rate`
        """
        if rate is None:
            raise ValueError("sampling_rate cannot be None")
        elif not hasattr(rate, "units"):
            raise ValueError("sampling_rate must have units")
        self._sampling_rate = rate

    # sampling_period attribute is handled as a property on underlying rate
    @property
    def sampling_period(self):
        """
        Interval between two samples.

        (1/:attr:`sampling_rate`)
        """
        return 1.0 / self.sampling_rate

    @sampling_period.setter
    def sampling_period(self, period):
        """
        Setter for :attr:`sampling_period`
        """
        if period is None:
            raise ValueError("sampling_period cannot be None")
        elif not hasattr(period, "units"):
            raise ValueError("sampling_period must have units")
        self.sampling_rate = 1.0 / period

    # t_start attribute is handled as a property so type checking can be done
    @property
    def t_start(self):
        """
        Time when signal begins.
        """
        return self._t_start

    @t_start.setter
    def t_start(self, start):
        """
        Setter for :attr:`t_start`
        """
        if start is None:
            raise ValueError("t_start cannot be None")
        self._t_start = start

    @property
    def duration(self):
        """
        Signal duration

        (:attr:`size` * :attr:`sampling_period`)
        """
        return self.shape[0] / self.sampling_rate

    @property
    def t_stop(self):
        """
        Time when signal ends.

        (:attr:`t_start` + :attr:`duration`)
        """
        return self.t_start + self.duration

    @property
    def times(self):
        """
        The time points of each sample of the signal

        (:attr:`t_start` + arange(:attr:`shape`)/:attr:`sampling_rate`)
        """
        return self.t_start + np.arange(self.shape[0]) / self.sampling_rate

    def __eq__(self, other):
        """
        Equality test (==)
        """
        if isinstance(other, AnalogSignal) and (
            self.t_start != other.t_start or self.sampling_rate != other.sampling_rate
        ):
            return False
        return super().__eq__(other)

    def _check_consistency(self, other):
        """
        Check if the attributes of another :class:`AnalogSignal`
        are compatible with this one.
        """
        if isinstance(other, AnalogSignal):
            for attr in "t_start", "sampling_rate":
                if getattr(self, attr) != getattr(other, attr):
                    raise ValueError(f"Inconsistent values of {attr}")  # how to handle name and annotations?

    def _repr_pretty_(self, pp, cycle):
        """
        Handle pretty-printing the :class:`AnalogSignal`.
        """
        pp.text(
            f"{self.__class__.__name__} with {self.shape[1]} channels of length "
            f"{self.shape[0]}; units {self.units.dimensionality.string}; datatype "
            f"{self.dtype}"
        )
        if self._has_repr_pretty_attrs_():
            pp.breakable()
            self._repr_pretty_attrs_(pp, cycle)

        def _pp(line):
            pp.breakable()
            with pp.group(indent=1):
                pp.text(line)

        _pp(f"sampling rate: {self.sampling_rate}")
        _pp(f"time: {self.t_start} to {self.t_stop}")

    def time_index(self, t):
        """Return the array index (or indices) corresponding to the time (or times) `t`"""
        i = (t - self.t_start) * self.sampling_rate
        i = np.rint(i.simplified.magnitude).astype(np.int64)
        return i

    def time_slice(self, t_start, t_stop):
        """
        Creates a new AnalogSignal corresponding to the time slice of the
        original AnalogSignal between times t_start, t_stop. Note, that for
        numerical stability reasons if t_start does not fall exactly on
        the time bins defined by the sampling_period it will be rounded to
        the nearest sampling bin. The time bin for t_stop will be chosen to
        make the duration of the resultant signal as close as possible to
        t_stop - t_start. This means that for a given duration, the size
        of the slice will always be the same.
        """

        # checking start time and transforming to start index
        if t_start is None:
            i = 0
            t_start = 0 * pq.s
        else:
            i = self.time_index(t_start)

        # checking stop time and transforming to stop index
        if t_stop is None:
            j = len(self)
        else:
            delta = (t_stop - t_start) * self.sampling_rate
            j = i + int(np.rint(delta.simplified.magnitude))

        if (i < 0) or (j > len(self)):
            raise ValueError(
                "t_start, t_stop have to be within the analog \
                              signal duration"
            )

        # Time slicing should create a deep copy of the object
        obj = deepcopy(self[i:j])

        obj.t_start = self.t_start + i * self.sampling_period

        return obj

    def time_shift(self, t_shift):
        """
        Shifts a :class:`AnalogSignal` to start at a new time.

        Parameters
        ----------
        t_shift: Quantity (time)
            Amount of time by which to shift the :class:`AnalogSignal`.

        Returns
        -------
        new_sig: :class:`AnalogSignal`
            New instance of a :class:`AnalogSignal` object starting at t_shift later than the
            original :class:`AnalogSignal` (the original :class:`AnalogSignal` is not modified).
        """
        new_sig = deepcopy(self)
        new_sig.t_start = new_sig.t_start + t_shift

        return new_sig

    def splice(self, signal, copy=False):
        """
        Replace part of the current signal by a new piece of signal.

        The new piece of signal will overwrite part of the current signal
        starting at the time given by the new piece's `t_start` attribute.

        The signal to be spliced in must have the same physical dimensions,
        sampling rate, and number of channels as the current signal and
        fit within it.

        If `copy` is False (the default), modify the current signal in place.
        If `copy` is True, return a new signal and leave the current one untouched.
        In this case, the new signal will not be linked to any parent objects.
        """
        if signal.t_start < self.t_start:
            raise ValueError("Cannot splice earlier than the start of the signal")
        if signal.t_stop > self.t_stop:
            raise ValueError("Splice extends beyond signal")
        if signal.sampling_rate != self.sampling_rate:
            raise ValueError("Sampling rates do not match")
        i = self.time_index(signal.t_start)
        j = i + signal.shape[0]
        if copy:
            new_signal = deepcopy(self)
            new_signal.segment = None
            new_signal[i:j, :] = signal
            return new_signal
        else:
            self[i:j, :] = signal
            return self

    def downsample(self, downsampling_factor, **kwargs):
        """
        Downsample the data of a signal.
        This method reduces the number of samples of the AnalogSignal to a fraction of the
        original number of samples, defined by `downsampling_factor`.
        This method is a wrapper of scipy.signal.decimate and accepts the same set of keyword
        arguments, except for specifying the axis of resampling, which is fixed to the first axis
        here.

        Parameters
        ----------
        downsampling_factor: int
            Factor used for decimation of samples. Scipy recommends to call decimate multiple times
            for downsampling factors higher than 13 when using IIR downsampling (default).

        Returns
        -------
        downsampled_signal: :class:`AnalogSignal`
            New instance of a :class:`AnalogSignal` object containing the resampled data points.
            The original :class:`AnalogSignal` is not modified.

        Notes
        -----
        For resampling the signal with a fixed number of samples, see `resample` method.
        """
        try:
            import scipy.signal
        except ImportError as err:
            raise ImportError("Decimating requires availability of scipy.signal")

        # Resampling is only permitted along the time axis (axis=0)
        if "axis" in kwargs:
            kwargs.pop("axis")

        downsampled_data = scipy.signal.decimate(self.magnitude, downsampling_factor, axis=0, **kwargs)
        downsampled_signal = self.duplicate_with_new_data(downsampled_data)

        # since the number of channels stays the same, we can also copy array annotations here
        downsampled_signal.array_annotations = self.array_annotations.copy()
        downsampled_signal.sampling_rate = self.sampling_rate / downsampling_factor

        return downsampled_signal

    def resample(self, sample_count, **kwargs):
        """
        Resample the data points of the signal.
        This method interpolates the signal and returns a new signal with a fixed number of
        samples defined by `sample_count`.
        This method is a wrapper of scipy.signal.resample and accepts the same set of keyword
        arguments, except for specifying the axis of resampling which is fixed to the first axis
        here, and the sample positions. .

        Parameters
        ----------
        sample_count: int
            Number of desired samples. The resulting signal starts at the same sample as the
            original and is sampled regularly.

        Returns
        -------
        resampled_signal: :class:`AnalogSignal`
            New instance of a :class:`AnalogSignal` object containing the resampled data points.
            The original :class:`AnalogSignal` is not modified.

        Notes
        -----
        For reducing the number of samples to a fraction of the original, see `downsample` method
        """

        try:
            import scipy.signal
        except ImportError as err:
            raise ImportError("Decimating requires availability of scipy.signal")

        # Resampling is only permitted along the time axis (axis=0)
        if "axis" in kwargs:
            kwargs.pop("axis")
        if "t" in kwargs:
            kwargs.pop("t")

        resampled_data, resampled_times = scipy.signal.resample(
            self.magnitude, sample_count, t=self.times, axis=0, **kwargs
        )

        resampled_signal = self.duplicate_with_new_data(resampled_data)
        resampled_signal.sampling_rate = (sample_count / self.shape[0]) * self.sampling_rate

        # since the number of channels stays the same, we can also copy array annotations here
        resampled_signal.array_annotations = self.array_annotations.copy()

        return resampled_signal

    def rectify(self, **kwargs):
        """
        Rectify the signal.
        This method rectifies the signal by taking the absolute value.
        This method is a wrapper of numpy.absolute() and accepts the same set of keyword
        arguments.

        Returns
        -------
        resampled_signal: :class:`AnalogSignal`
            New instance of a :class:`AnalogSignal` object containing the rectified data points.
            The original :class:`AnalogSignal` is not modified.

        """

        # Use numpy to get the absolute value of the signal
        rectified_data = np.absolute(self.magnitude, **kwargs)

        rectified_signal = self.duplicate_with_new_data(rectified_data)

        # the sampling rate stays constant
        rectified_signal.sampling_rate = self.sampling_rate

        # since the number of channels stays the same, we can also copy array annotations here
        rectified_signal.array_annotations = self.array_annotations.copy()

        return rectified_signal

    def concatenate(self, *signals, overwrite=False, padding=False):
        """
        Concatenate multiple neo.AnalogSignal objects across time.

        Units, sampling_rate and number of signal traces must be the same
        for all signals. Otherwise a ValueError is raised.
        Note that timestamps of concatenated signals might shift in oder to
        align the sampling times of all signals.

        Parameters
        ----------
        signals: neo.AnalogSignal objects
            AnalogSignals that will be concatenated
        overwrite : bool
            If True, samples of the earlier (lower index in `signals`)
            signals are overwritten by that of later (higher index in `signals`)
            signals.
            If False, samples of the later are overwritten by earlier signal.
            Default: False
        padding : bool, scalar quantity
            Sampling values to use as padding in case signals do not overlap.
            If False, do not apply padding. Signals have to align or
            overlap. If True, signals will be padded using
            np.NaN as pad values. If a scalar quantity is provided, this
            will be used for padding. The other signal is moved
            forward in time by maximum one sampling period to
            align the sampling times of both signals.
            Default: False

        Returns
        -------
        signal: neo.AnalogSignal
            concatenated output signal
        """

        # Sanity of inputs
        if not hasattr(signals, "__iter__"):
            raise TypeError("signals must be iterable")
        if not all([isinstance(a, AnalogSignal) for a in signals]):
            raise TypeError("Entries of anasiglist have to be of type neo.AnalogSignal")
        if len(signals) == 0:
            return self

        signals = [self] + list(signals)

        # Check required common attributes: units, sampling_rate and shape[-1]
        shared_attributes = ["units", "sampling_rate"]
        attribute_values = [tuple((getattr(anasig, attr) for attr in shared_attributes)) for anasig in signals]
        # add shape dimensions that do not relate to time
        attribute_values = [(attribute_values[i] + (signals[i].shape[1:],)) for i in range(len(signals))]
        if not all([attrs == attribute_values[0] for attrs in attribute_values]):
            raise MergeError(f"AnalogSignals have to share {shared_attributes} attributes to be concatenated.")
        units, sr, shape = attribute_values[0]

        # find gaps between Analogsignals
        combined_time_ranges = self._concatenate_time_ranges([(s.t_start, s.t_stop) for s in signals])
        missing_time_ranges = self._invert_time_ranges(combined_time_ranges)
        if len(missing_time_ranges):
            diffs = np.diff(np.asarray(missing_time_ranges), axis=1)
        else:
            diffs = []

        if padding is False and any(diffs > signals[0].sampling_period):
            raise MergeError(
                f"Signals are not continuous. Can not concatenate signals with gaps. "
                f"Please provide a padding value."
            )
        if padding is not False:
            logger.warning(f"Signals will be padded using {padding}.")
            if padding is True:
                padding = np.nan * units
            if isinstance(padding, pq.Quantity):
                padding = padding.rescale(units).magnitude
            else:
                raise MergeError(
                    "Invalid type of padding value. Please provide a bool value " "or a quantities object."
                )

        t_start = min([a.t_start for a in signals])
        t_stop = max([a.t_stop for a in signals])
        n_samples = int(np.rint(((t_stop - t_start) * sr).rescale("dimensionless").magnitude))
        shape = (n_samples,) + shape

        # Collect attributes and annotations across all concatenated signals
        kwargs = {}
        common_annotations = signals[0].annotations
        common_array_annotations = signals[0].array_annotations
        for anasig in signals[1:]:
            common_annotations = intersect_annotations(common_annotations, anasig.annotations)
            common_array_annotations = intersect_annotations(common_array_annotations, anasig.array_annotations)

        kwargs["annotations"] = common_annotations
        kwargs["array_annotations"] = common_array_annotations

        for name in ("name", "description", "file_origin"):
            attr = [getattr(s, name) for s in signals]
            if all([a == attr[0] for a in attr]):
                kwargs[name] = attr[0]
            else:
                kwargs[name] = f"concatenation ({attr})"

        conc_signal = AnalogSignal(
            np.full(shape=shape, fill_value=padding, dtype=signals[0].dtype),
            sampling_rate=sr,
            t_start=t_start,
            units=units,
            **kwargs,
        )

        if not overwrite:
            signals = signals[::-1]
        while len(signals) > 0:
            conc_signal.splice(signals.pop(0), copy=False)

        return conc_signal

    def _concatenate_time_ranges(self, time_ranges):
        time_ranges = sorted(time_ranges)
        new_ranges = time_ranges[:1]
        for t_start, t_stop in time_ranges[1:]:
            # time range are non continuous -> define new range
            if t_start > new_ranges[-1][1]:
                new_ranges.append((t_start, t_stop))
            # time range is continuous -> extend time range
            elif t_stop > new_ranges[-1][1]:
                new_ranges[-1] = (new_ranges[-1][0], t_stop)
        return new_ranges

    def _invert_time_ranges(self, time_ranges):
        i = 0
        new_ranges = []
        while i < len(time_ranges) - 1:
            new_ranges.append((time_ranges[i][1], time_ranges[i + 1][0]))
            i += 1
        return new_ranges



import unittest

from neo.rawio.axonrawio import AxonRawIO

from neo.test.rawiotest.common_rawio_test import BaseTestRawIO


class TestAxonRawIO(
    BaseTestRawIO,
    unittest.TestCase,
):
    rawioclass = AxonRawIO
    entities_to_test = [
        "axon/File_axon_1.abf",  # V2.0
        "axon/File_axon_2.abf",  # V1.8
        "axon/File_axon_3.abf",  # V1.8
        "axon/File_axon_4.abf",  # 2.0
        "axon/File_axon_5.abf",  # V.20
        "axon/File_axon_6.abf",  # V.20
        "axon/File_axon_7.abf",  # V2.6
        "axon/test_file_edr3.abf",  # EDR3
    ]
    entities_to_download = ["axon"]

    def test_read_raw_protocol(self):
        reader = AxonRawIO(filename=self.get_local_path("axon/File_axon_7.abf"))
        reader.parse_header()

        reader.read_raw_protocol()


if __name__ == "__main__":
    unittest.main()



"""
Tests of neo.io.asciisignalio
"""

import os
import unittest
import json
import csv
import numpy as np
import quantities as pq
from numpy.testing import assert_array_almost_equal, assert_array_equal
from neo.io import AsciiSignalIO
from neo.test.iotest.common_io_test import BaseTestIO
from neo.core import AnalogSignal, Segment, Block


class TestAsciiSignalIOWithTestFiles(BaseTestIO, unittest.TestCase):
    ioclass = AsciiSignalIO
    entities_to_download = ["asciisignal"]
    entities_to_test = [
        "asciisignal/File_asciisignal_2.txt",
        "asciisignal/File_asciisignal_3.txt",
    ]


class TestAsciiSignalIO(unittest.TestCase):

    def test_genfromtxt_expect_success(self):
        sample_data = np.random.uniform(size=(200, 3))
        filename = "test_genfromtxt_expect_success.txt"
        np.savetxt(filename, sample_data, delimiter=" ")
        sampling_rate = 1 * pq.kHz
        io = AsciiSignalIO(filename, sampling_rate=sampling_rate, delimiter=" ", units="mV", method="genfromtxt")
        block = io.read_block()

        signal1 = block.segments[0].analogsignals[1]
        assert_array_almost_equal(signal1.reshape(-1).magnitude, sample_data[:, 1], decimal=6)
        self.assertEqual(len(block.segments[0].analogsignals), 3)
        self.assertEqual(signal1.t_stop, sample_data.shape[0] / sampling_rate)
        self.assertEqual(signal1.units, pq.mV)

        os.remove(filename)

    # test_genfromtxt_expect_failure

    def test_csv_expect_success(self):
        filename = "test_csv_expect_success.csv"
        sample_data = [
            (-65, -65, -65, 0.5),
            (-64.8, -64.5, -64.0, 0.6),
            (-64.6, -64.2, -77.0, 0.7),
            (-64.3, -64.0, -99.9, 0.8),
        ]
        with open(filename, "w", newline="") as csvfile:
            writer = csv.writer(csvfile, delimiter=",", quotechar="|", quoting=csv.QUOTE_MINIMAL)
            for row in sample_data:
                writer.writerow(row)

        usecols = (0, 1, 3)
        timecolumn = 2
        io = AsciiSignalIO(
            filename,
            usecols=usecols,
            timecolumn=timecolumn,
            # note that timecolumn applies to the remaining columns
            # after applying usecols
            time_units="ms",
            delimiter=",",
            units="mV",
            method="csv",
            signal_group_mode="all-in-one",
            t_start=0.5,
        )

        block = io.read_block()
        signal = block.segments[0].analogsignals[0]
        self.assertEqual(signal.shape, (4, 2))  # two columns remaining after usecols
        # and timecolumn applied
        assert_array_almost_equal(signal[:, 1].reshape(-1).magnitude, np.array(sample_data)[:, 1], decimal=5)
        self.assertAlmostEqual(signal.sampling_period, 0.1 * pq.ms)

        expected_channel_index = list(usecols)
        # remove time column as it is not loaded as signal channel
        expected_channel_index.pop(timecolumn)
        assert_array_equal(expected_channel_index, signal.array_annotations["channel_index"])

        os.remove(filename)

    # test_csv_expect_failure

    def test_homemade_expect_success(self):
        filename = "test_homemade_expect_success.txt"
        sample_data = [
            (-65, -65, -65, 0.5),
            (-64.8, -64.5, -64.0, 0.6),
            (-64.6, -64.2, -77.0, 0.7),
            (-64.3, -64.0, -99.9, 0.8),
        ]
        with open(filename, "w") as datafile:
            datafile.write("# a comment\n")
            for row in sample_data:
                datafile.write("\t ".join(map(str, row)) + "\t\n")

        io = AsciiSignalIO(
            filename,
            usecols=(0, 1, 3),
            timecolumn=2,
            skiprows=1,
            time_units="ms",
            delimiter="\t",
            units="mV",
            method="homemade",
            signal_group_mode="all-in-one",
            t_start=0.5,
        )

        block = io.read_block()
        signal = block.segments[0].analogsignals[0]
        self.assertEqual(signal.shape, (4, 2))  # two columns remaining after usecols
        # and timecolumn applied
        assert_array_almost_equal(signal[:, 1].reshape(-1).magnitude, np.array(sample_data)[:, 1], decimal=5)
        self.assertAlmostEqual(signal.sampling_period, 0.1 * pq.ms)

        os.remove(filename)

    # test_homemade_expect_failure

    def test_callable_expect_success(self):
        sample_data = np.random.uniform(size=(200, 3))
        filename = "test_genfromtxt_expect_success.txt"
        np.savetxt(filename, sample_data, delimiter=" ")
        sampling_rate = 1 * pq.kHz

        def reader(filename, comment_rows):
            return np.genfromtxt(filename, delimiter=" ", usecols=None, skip_header=comment_rows or 0, dtype="f")

        io = AsciiSignalIO(filename, sampling_rate=sampling_rate, delimiter=" ", units="mV", method=reader)
        block = io.read_block()

        signal1 = block.segments[0].analogsignals[1]
        assert_array_almost_equal(signal1.reshape(-1).magnitude, sample_data[:, 1], decimal=6)
        self.assertEqual(len(block.segments[0].analogsignals), 3)
        self.assertEqual(signal1.t_stop, sample_data.shape[0] / sampling_rate)
        self.assertEqual(signal1.units, pq.mV)

        os.remove(filename)

    def test_skiprows(self):
        filename = "test_skiprows.txt"
        sample_data = [
            (-65, -65, -65, 0.5),
            (-64.8, -64.5, -64.0, 0.6),
            (-64.6, -64.2, -77.0, 0.7),
            (-64.3, -64.0, -99.9, 0.8),
            (-66, -64.3, -61.0, 1.4),
        ]
        with open(filename, "w") as datafile:
            for row in sample_data:
                datafile.write("\t ".join(map(str, row)) + "\t\n")
        io = AsciiSignalIO(filename, skiprows=1)
        block = io.read_block()
        signal = block.segments[0].analogsignals[0]
        self.assertEqual(signal.shape, (4, 1))
        self.assertEqual(len(block.segments[0].analogsignals), 5)
        self.assertEqual(len(block.segments[0].analogsignals[0]), 4)
        self.assertEqual(signal.units, pq.V)
        assert_array_equal(signal.times, [0.0, 1.0, 2.0, 3.0] * pq.s)
        assert_array_equal(signal.times.magnitude, [0.0, 1.0, 2.0, 3.0])
        assert_array_equal(signal[0].magnitude, -64.8)
        assert_array_equal(signal[1].magnitude, -64.6)
        assert_array_equal(signal[2].magnitude, -64.3)
        assert_array_equal(signal[3].magnitude, -66)
        assert_array_almost_equal(np.asarray(signal).flatten(), np.array([-64.8, -64.6, -64.3, -66]), decimal=5)

        os.remove(filename)

    def test_usecols(self):
        filename = "test_usecols.txt"
        sample_data = [
            (-65, -65, -65, 0.5),
            (-64.8, -64.5, -64.0, 0.6),
            (-64.6, -64.2, -77.0, 0.7),
            (-64.3, -64.0, -99.9, 0.8),
            (-66, -64.3, -61.0, 1.4),
        ]
        with open(filename, "w") as datafile:
            for row in sample_data:
                datafile.write("\t ".join(map(str, row)) + "\t\n")
        io = AsciiSignalIO(filename, usecols=3)
        block = io.read_block()
        signal = block.segments[0].analogsignals[0]
        self.assertEqual(signal.shape, (5, 1))
        self.assertEqual(len(block.segments[0].analogsignals[0]), 5)
        self.assertEqual(len(block.segments[0].analogsignals), 1)
        self.assertEqual(signal.units, pq.V)
        assert_array_equal(signal.times, [0.0, 1.0, 2.0, 3.0, 4.0] * pq.s)
        assert_array_equal(signal.times.magnitude, [0.0, 1.0, 2.0, 3.0, 4.0])
        assert_array_equal(signal[0].magnitude, 0.5)
        assert_array_equal(signal[1].magnitude, 0.6)
        assert_array_equal(signal[2].magnitude, 0.7)
        assert_array_equal(signal[3].magnitude, 0.8)
        assert_array_equal(signal[4].magnitude, 1.4)
        assert_array_almost_equal(np.asarray(signal).flatten(), np.array([0.5, 0.6, 0.7, 0.8, 1.4]), decimal=5)

        os.remove(filename)

    # test skip comments
    def test_skip_comments(self):

        sample_data = [
            ("INFORMATION"),
            ("20170119_cell1_CA"),
            (""),
            (""),
            (""),
            (""),
            (""),
            (),
            ("Waveform"),
            ("Unit"),
            (),
            (-0.02121),
            (-0.01740),
            (-0.00671),
            (-0.00885),
            (-0.00092),
            (-0.01373),
            (-0.01175),
            (-0.00504),
            (-0.01358),
            (-0.01526),
            (-0.00244),
            (2.26852),
            ("CHANNEL" "5"),
            ("Evt+-"),
            (""),
            (""),
        ]
        sampling_rate = 1 * pq.kHz
        filename = "test_skip_comments.txt"
        with open(filename, "w") as datafile:
            for row in sample_data:
                datafile.write(str(float))

        io = AsciiSignalIO(filename, sampling_rate=sampling_rate, delimiter=" ", units="mV", method="genfromtxt")
        block = io.read_block()
        signal1 = block.segments[0].analogsignals[0]

        self.assertEqual(len(block.segments[0].analogsignals), 1)
        self.assertEqual(signal1.units, pq.mV)
        self.assertEqual(len(block.segments[0].analogsignals), 1)

        os.remove(filename)

    def test_timecolumn(self):
        sample_data = np.random.uniform(size=(200, 3))
        sampling_period = 0.5
        time_data = sampling_period * np.arange(sample_data.shape[0])
        combined_data = np.hstack((sample_data, time_data[:, np.newaxis]))
        filename = "test_multichannel.txt"
        np.savetxt(filename, combined_data, delimiter=" ")
        io = AsciiSignalIO(
            filename,
            delimiter=" ",
            units="mV",
            method="genfromtxt",
            timecolumn=-1,
            time_units="ms",
            signal_group_mode="split-all",
        )
        block = io.read_block()

        signal1 = block.segments[0].analogsignals[1]
        assert_array_almost_equal(signal1.reshape(-1).magnitude, sample_data[:, 1], decimal=6)
        self.assertEqual(signal1.sampling_period, sampling_period * pq.ms)
        self.assertEqual(len(block.segments[0].analogsignals), 3)
        self.assertEqual(signal1.t_stop, sample_data.shape[0] * sampling_period * pq.ms)
        self.assertEqual(signal1.units, pq.mV)

        os.remove(filename)

    def test_multichannel(self):
        sample_data = np.random.uniform(size=(200, 3))
        filename = "test_multichannel.txt"
        np.savetxt(filename, sample_data, delimiter=" ")
        sampling_rate = 1 * pq.kHz
        io = AsciiSignalIO(
            filename,
            sampling_rate=sampling_rate,
            delimiter=" ",
            units="mV",
            method="genfromtxt",
            signal_group_mode="all-in-one",
        )
        block = io.read_block()

        signal = block.segments[0].analogsignals[0]
        assert_array_almost_equal(signal.magnitude, sample_data, decimal=6)
        self.assertEqual(len(block.segments[0].analogsignals), 1)
        self.assertEqual(signal.t_stop, sample_data.shape[0] / sampling_rate)
        self.assertEqual(signal.units, pq.mV)

        os.remove(filename)

    def test_multichannel_with_timecolumn(self):
        sample_data = np.random.uniform(size=(200, 3))
        sampling_period = 0.5
        time_data = sampling_period * np.arange(sample_data.shape[0])
        combined_data = np.hstack((time_data[:, np.newaxis], sample_data))
        filename = "test_multichannel.txt"
        np.savetxt(filename, combined_data, delimiter=" ")
        io = AsciiSignalIO(
            filename,
            delimiter=" ",
            units="mV",
            method="genfromtxt",
            timecolumn=0,
            time_units="ms",
            signal_group_mode="all-in-one",
        )
        block = io.read_block()

        signal = block.segments[0].analogsignals[0]
        assert_array_almost_equal(signal.magnitude, sample_data, decimal=6)
        self.assertEqual(signal.sampling_period, sampling_period * pq.ms)
        self.assertEqual(len(block.segments[0].analogsignals), 1)
        self.assertEqual(signal.t_stop, sample_data.shape[0] * sampling_period * pq.ms)
        self.assertEqual(signal.units, pq.mV)

        os.remove(filename)

    def test_multichannel_with_negative_timecolumn(self):
        sample_data = np.random.uniform(size=(200, 3))
        sampling_period = 0.5
        time_data = sampling_period * np.arange(sample_data.shape[0])
        combined_data = np.hstack((sample_data, time_data[:, np.newaxis]))
        filename = "test_multichannel.txt"
        np.savetxt(filename, combined_data, delimiter=" ")
        io = AsciiSignalIO(
            filename,
            delimiter=" ",
            units="mV",
            method="genfromtxt",
            timecolumn=-1,
            time_units="ms",
            signal_group_mode="all-in-one",
        )
        block = io.read_block()

        signal = block.segments[0].analogsignals[0]
        assert_array_almost_equal(signal.magnitude, sample_data, decimal=6)
        self.assertEqual(signal.sampling_period, sampling_period * pq.ms)
        self.assertEqual(len(block.segments[0].analogsignals), 1)
        self.assertEqual(signal.t_stop, sample_data.shape[0] * sampling_period * pq.ms)
        self.assertEqual(signal.units, pq.mV)

        os.remove(filename)

    # test write without timecolumn
    def test_write_without_timecolumn(self):
        sample_data = np.random.uniform(size=(2, 3))
        sampling_period = 1
        time_data = sampling_period * np.arange(sample_data.shape[0])
        combined_data = np.hstack((time_data[:, np.newaxis], sample_data))
        filename = "test_write_without_timecolumn.txt"
        np.savetxt(filename, combined_data, delimiter=" ")

        signal1 = AnalogSignal(sample_data, units="V", sampling_rate=1 * pq.Hz)
        seg1 = Segment()
        block1 = Block()
        seg1.analogsignals.append(signal1)
        block1.segments.append(seg1)

        iow = AsciiSignalIO(filename, method="genfromtxt", time_units="ms", signal_group_mode="all-in-one")
        iow.write_block(block1)

        ior = AsciiSignalIO(filename)
        block2 = ior.read_block()

        assert len(block2.segments[0].analogsignals) == 3
        signal2 = block2.segments[0].analogsignals[1]

        assert_array_almost_equal(signal1.magnitude[:, 1], signal2.magnitude.reshape(-1), decimal=7)
        self.assertEqual(signal1.units, signal2.units)
        self.assertEqual(signal1.sampling_rate, signal2.sampling_rate)
        assert_array_equal(signal1.times, signal2.times)

        assert_array_almost_equal(signal2.magnitude.reshape(-1), sample_data[:, 1], decimal=6)

        self.assertEqual(signal2.sampling_period, sampling_period * pq.s)
        self.assertEqual(len(block2.segments[0].analogsignals), 3)
        self.assertEqual(signal2.t_stop, sample_data.shape[0] * sampling_period * pq.s)
        self.assertEqual(signal2.units, pq.V)

        os.remove(filename)

    # test write with timecolumn
    def test_write_with_timecolumn(self):
        sample_data = np.random.uniform(size=(200, 3))
        sampling_period = 1
        time_data = sampling_period * np.arange(sample_data.shape[0])
        combined_data = np.hstack((time_data[:, np.newaxis], sample_data))
        filename = "test_write_with_timecolumn.txt"
        np.savetxt(filename, combined_data, delimiter=" ")

        signal1 = AnalogSignal(sample_data, units="V", sampling_rate=1 * pq.Hz)
        seg1 = Segment()
        block1 = Block()
        seg1.analogsignals.append(signal1)
        block1.segments.append(seg1)

        iow = AsciiSignalIO(
            filename, method="genfromtxt", timecolumn=0, time_units="ms", signal_group_mode="all-in-one"
        )
        iow.write_block(block1)

        ior = AsciiSignalIO(filename)
        block2 = ior.read_block()

        assert len(block2.segments[0].analogsignals) == 4
        signal2 = block2.segments[0].analogsignals[1]

        assert_array_almost_equal(signal1.magnitude[:, 0], signal2.magnitude.reshape(-1), decimal=7)
        self.assertEqual(signal1.units, signal2.units)
        self.assertEqual(signal1.sampling_rate, signal2.sampling_rate)
        assert_array_equal(signal1.times, signal2.times)
        assert_array_almost_equal(signal2.magnitude[:, 0], sample_data[:, 0], decimal=6)
        self.assertEqual(signal2.sampling_period, sampling_period * pq.s)
        self.assertEqual(len(block2.segments[0].analogsignals), 4)
        self.assertEqual(signal2.t_stop, sample_data.shape[0] * sampling_period * pq.s)
        self.assertEqual(signal2.units, pq.V)

        os.remove(filename)

    # test write with units/timeunits different from those of signal
    def test_write_with_timeunits_different_from_those_of_signal(self):
        sample_data = np.random.uniform(size=(3, 3))
        sampling_period = 1
        time_data = sampling_period * np.arange(sample_data.shape[0])
        combined_data = np.hstack((time_data[:, np.newaxis], sample_data))
        filename = "test_write_with_timeunits_different_from_those_of_signal.txt"
        np.savetxt(filename, combined_data, delimiter=" ")

        signal1 = AnalogSignal(sample_data, units="A", sampling_rate=2 * pq.Hz, sampling_period=0.5)
        seg1 = Segment()
        block1 = Block()
        seg1.analogsignals.append(signal1)
        block1.segments.append(seg1)

        iow = AsciiSignalIO(
            filename, method="genfromtxt", timecolumn=0, units="mV", time_units="ms", signal_group_mode="all-in-one"
        )
        def_units = iow.units
        def_time_units = iow.time_units
        iow.write_block(block1)

        ior = AsciiSignalIO(filename)
        block2 = ior.read_block()

        assert len(block2.segments[0].analogsignals) == 4
        signal2 = block2.segments[0].analogsignals[1]

        self.assertTrue(signal1.units, def_units)
        self.assertEqual(signal1.units, pq.A)
        self.assertEqual(iow.units, pq.A)
        self.assertEqual(signal1.units, iow.units)
        self.assertTrue(def_units, iow.units)

        self.assertTrue(signal1.units, ior.time_units)
        self.assertEqual(iow.time_units, ior.time_units)
        self.assertTrue(def_time_units, iow.time_units)
        self.assertEqual(iow.time_units, pq.s)
        self.assertEqual(ior.time_units, pq.s)

        os.remove(filename)

    def test_read_with_json_metadata(self):
        sample_data = np.random.uniform(size=(200, 3))
        filename = "test_read_with_json_metadata.txt"
        metadata_filename = "test_read_with_json_metadata_about.json"
        np.savetxt(filename, sample_data, delimiter=" ")

        metadata = {
            "filename": filename,
            "delimiter": " ",
            "timecolumn": None,
            "units": "mV",
            "time_units": "ms",
            "sampling_rate": {"value": 1.0, "units": "kHz"},
            "method": "genfromtxt",
            "signal_group_mode": "split-all",
        }
        with open(metadata_filename, "w") as fp:
            json.dump(metadata, fp)

        expected_sampling_rate = 1 * pq.kHz
        io = AsciiSignalIO(filename)
        block = io.read_block()

        signal1 = block.segments[0].analogsignals[1]
        assert_array_almost_equal(signal1.reshape(-1).magnitude, sample_data[:, 1], decimal=6)
        self.assertEqual(len(block.segments[0].analogsignals), 3)
        self.assertEqual(signal1.sampling_rate, expected_sampling_rate)
        self.assertEqual(signal1.t_stop, sample_data.shape[0] / signal1.sampling_rate)
        self.assertEqual(signal1.units, pq.mV)

        os.remove(filename)
        os.remove(metadata_filename)

    def test_roundtrip_with_json_metadata(self):
        sample_data = np.random.uniform(size=(200, 3))
        filename = "test_roundtrip_with_json_metadata.txt"
        metadata_filename = "test_roundtrip_with_json_metadata_about.json"
        signal1 = AnalogSignal(sample_data, units="pA", sampling_rate=2 * pq.kHz)
        seg1 = Segment()
        block1 = Block()
        seg1.analogsignals.append(signal1)
        block1.segments.append(seg1)

        iow = AsciiSignalIO(filename, metadata_filename=metadata_filename)
        iow.write_block(block1)
        self.assertTrue(os.path.exists(metadata_filename))

        ior = AsciiSignalIO(filename)
        block2 = ior.read_block()
        assert len(block2.segments[0].analogsignals) == 3
        signal2 = block2.segments[0].analogsignals[1]

        assert_array_almost_equal(signal1.magnitude[:, 1], signal2.magnitude.reshape(-1), decimal=7)
        self.assertEqual(signal1.units, signal2.units)
        self.assertEqual(signal1.sampling_rate, signal2.sampling_rate)
        assert_array_equal(signal1.times, signal2.times)

        os.remove(filename)
        os.remove(metadata_filename)

    def test_genfromtxt_irregular_expect_success(self):
        sample_data = np.random.uniform(size=(200, 3))
        sample_data[:, 0] = np.sort(sample_data[:, 0])  # make column 0 the time column
        filename = "test_genfromtxt_irregular_expect_success.txt"
        np.savetxt(filename, sample_data, delimiter=" ")

        io = AsciiSignalIO(
            filename, delimiter=" ", timecolumn=0, units="mV", method="genfromtxt", signal_group_mode="split-all"
        )
        block = io.read_block()

        signal1 = block.segments[0].irregularlysampledsignals[1]
        assert_array_almost_equal(signal1.reshape(-1).magnitude, sample_data[:, 2], decimal=6)
        self.assertEqual(len(block.segments[0].analogsignals), 0)
        self.assertEqual(len(block.segments[0].irregularlysampledsignals), 2)
        self.assertEqual(signal1.units, pq.mV)

        os.remove(filename)

    def test_irregular_multichannel(self):
        sample_data = np.random.uniform(size=(200, 3))
        sample_data[:, 0] = np.sort(sample_data[:, 0])  # make column 0 the time column
        filename = "test_irregular_multichannel.txt"
        np.savetxt(filename, sample_data, delimiter=" ")

        io = AsciiSignalIO(
            filename, delimiter=" ", timecolumn=0, units="mV", method="genfromtxt", signal_group_mode="all-in-one"
        )
        block = io.read_block()

        signal = block.segments[0].irregularlysampledsignals[0]
        assert_array_almost_equal(signal.magnitude, sample_data[:, 1:3], decimal=6)
        self.assertEqual(len(block.segments[0].analogsignals), 0)
        self.assertEqual(len(block.segments[0].irregularlysampledsignals), 1)
        self.assertEqual(signal.shape, (200, 2))
        self.assertEqual(signal.units, pq.mV)

        os.remove(filename)


if __name__ == "__main__":
    unittest.main()



"""
Tests of the neo.core.baseneo.BaseNeo class and related functions
"""

from datetime import datetime, date, time, timedelta
from decimal import Decimal
from fractions import Fraction
import sys

import unittest

import numpy as np
import quantities as pq

try:
    from IPython.lib.pretty import pretty
except ImportError as err:
    HAVE_IPYTHON = False
else:
    HAVE_IPYTHON = True

from neo.core.baseneo import BaseNeo, _check_annotations, merge_annotations, merge_annotation, intersect_annotations
from neo.test.tools import assert_arrays_equal


class Test_check_annotations(unittest.TestCase):
    """
    TestCase to make sure _check_annotations works
    """

    def setUp(self):
        self.values = [
            1,
            2.2,
            3 + 2j,
            "test",
            r"test",
            b"test",
            None,
            datetime(year=2008, month=12, day=3, hour=10, minute=4),
            timedelta(weeks=2, days=7, hours=18, minutes=28, seconds=18, milliseconds=28, microseconds=45),
            time(hour=10, minute=4),
            Decimal("3.14"),
            Fraction(13, 21),
            np.array([1.1, 1.2, 1.3]),
            np.array([1, 2, 3]),
            np.array("test", dtype="S"),
            np.array([True, False]),
        ]

    def test__check_annotations__invalid_ValueError(self):
        value = set()
        self.assertRaises(ValueError, _check_annotations, value)

    def test__check_annotations__invalid_dtype_ValueError(self):
        value = np.array([], dtype="O")
        self.assertRaises(ValueError, _check_annotations, value)

    def test__check_annotations__valid_dtypes(self):
        for value in self.values:
            _check_annotations(value)

    def test__check_annotations__list(self):
        _check_annotations(self.values)

    def test__check_annotations__tuple(self):
        _check_annotations(tuple(self.values))
        _check_annotations((self.values, self.values))

    def test__check_annotations__dict(self):
        names = [f"value{i}" for i in range(len(self.values))]
        values = dict(zip(names, self.values))
        _check_annotations(values)


class TestBaseNeo(unittest.TestCase):
    """
    TestCase to make sure basic initialization and methods work
    """

    def assertDictContainsSubset(self, a, b):
        """Checks that all the key/value pairs in a exist in b"""
        # the implementation in unittest.TestCase has been deprecated
        assert set(a.items()).issubset(set(b.items()))

    def test_init(self):
        """test to make sure initialization works properly"""
        base = BaseNeo(name="a base", description="this is a test")
        self.assertEqual(base.name, "a base")
        self.assertEqual(base.description, "this is a test")
        self.assertEqual(base.file_origin, None)

    def test_annotate(self):
        """test to make sure annotation works properly"""
        base = BaseNeo()
        base.annotate(test1=1, test2=1)
        result1 = {"test1": 1, "test2": 1}

        self.assertDictEqual(result1, base.annotations)

        base.annotate(test3=2, test4=3)
        result2 = {"test3": 2, "test4": 3}
        result2a = dict(list(result1.items()) + list(result2.items()))

        self.assertDictContainsSubset(result1, base.annotations)
        self.assertDictContainsSubset(result2, base.annotations)
        self.assertDictEqual(result2a, base.annotations)

        base.annotate(test1=5, test2=8)
        result3 = {"test1": 5, "test2": 8}
        result3a = dict(list(result3.items()) + list(result2.items()))

        self.assertDictContainsSubset(result2, base.annotations)
        self.assertDictContainsSubset(result3, base.annotations)
        self.assertDictEqual(result3a, base.annotations)

        self.assertNotEqual(base.annotations["test1"], result1["test1"])
        self.assertNotEqual(base.annotations["test2"], result1["test2"])

    def test__children(self):
        base = BaseNeo()

        self.assertEqual(base._parent_objects, ())

        self.assertEqual(base._parent_containers, ())

        self.assertEqual(base._parent_objects, ())
        self.assertEqual(base._parent_containers, ())

        self.assertEqual(base.parents, ())


class Test_BaseNeo_merge_annotations_merge(unittest.TestCase):
    """
    TestCase to make sure merge_annotations and merge methods work
    """

    def setUp(self):
        self.name1 = "a base 1"
        self.name2 = "a base 2"
        self.name3 = "a base 3"
        self.description1 = "this is a test 1"
        self.description2 = "this is a test 2"
        self.description3 = "this is a test 3"
        self.base1 = BaseNeo(name=self.name1, description=self.description1)
        self.base2 = BaseNeo(name=self.name2, description=self.description2)
        self.base3 = BaseNeo(name=self.name3, description=self.description3)

    def test_merge_annotations__dict(self):
        self.base1.annotations = {
            "val0": "val0",
            "val1": 1,
            "val2": 2.2,
            "val3": "test1",
            "val4": [0.4],
            "val5": {0: 0, 1: {0: 0}},
            "val6": np.array([0, 1, 2]),
        }
        self.base2.annotations = {
            "val2": 2.2,
            "val3": "test2",
            "val4": [4, 4.4],
            "val5": {1: {1: 1}, 2: 2},
            "val6": np.array([4, 5, 6]),
            "val7": True,
        }

        ann1 = self.base1.annotations
        ann2 = self.base2.annotations
        ann1c = self.base1.annotations.copy()
        ann2c = self.base2.annotations.copy()

        targ = {
            "val0": "val0",
            "val1": 1,
            "val2": 2.2,
            "val3": "test1;test2",
            "val4": [0.4, 4, 4.4],
            "val5": {0: 0, 1: {0: 0, 1: 1}, 2: 2},
            "val7": True,
        }

        self.base1.merge_annotations(self.base2)

        val6t = np.array([0, 1, 2, 4, 5, 6])
        val61 = ann1.pop("val6")
        val61c = ann1c.pop("val6")
        val62 = ann2.pop("val6")
        val62c = ann2c.pop("val6")

        self.assertEqual(ann1, self.base1.annotations)
        self.assertNotEqual(ann1c, self.base1.annotations)
        self.assertEqual(ann2c, self.base2.annotations)
        self.assertEqual(targ, self.base1.annotations)

        assert_arrays_equal(val61, val6t)
        self.assertRaises(AssertionError, assert_arrays_equal, val61c, val6t)
        assert_arrays_equal(val62, val62c)

        self.assertEqual(self.name1, self.base1.name)
        self.assertEqual(self.name2, self.base2.name)
        self.assertEqual(self.description1, self.base1.description)
        self.assertEqual(self.description2, self.base2.description)

    def test_merge_multiple_annotations__dict(self):
        self.base1.annotations = {
            "val0": "val0",
            "val1": 1,
            "val2": 2.2,
            "val3": "test1",
            "val4": [0.4],
            "val5": {0: 0, 1: {0: 0}},
            "val6": np.array([0, 1, 2]),
        }
        self.base2.annotations = {
            "val2": 2.2,
            "val3": "test2",
            "val4": [4, 4.4],
            "val5": {1: {1: 1}, 2: 2},
            "val6": np.array([4, 5, 6]),
            "val7": True,
        }
        self.base3.annotations = {
            "val2": 2.2,
            "val3": "test3",
            "val4": [44],
            "val5": {1: {2: 2}, 2: 2, 3: 3},
            "val6": np.array([8, 9, 10]),
            "val8": False,
        }

        ann1 = self.base1.annotations
        ann2 = self.base2.annotations
        ann3 = self.base3.annotations
        ann1c = self.base1.annotations.copy()
        ann2c = self.base2.annotations.copy()
        ann3c = self.base3.annotations.copy()

        targ = {
            "val0": "val0",
            "val1": 1,
            "val2": 2.2,
            "val3": "test1;test2;test3",
            "val4": [0.4, 4, 4.4, 44],
            "val5": {0: 0, 1: {0: 0, 1: 1, 2: 2}, 2: 2, 3: 3},
            "val7": True,
            "val8": False,
        }

        self.base1.merge_annotations(self.base2, self.base3)

        val6t = np.array([0, 1, 2, 4, 5, 6, 8, 9, 10])
        val61 = ann1.pop("val6")
        val61c = ann1c.pop("val6")
        val62 = ann2.pop("val6")
        val62c = ann2c.pop("val6")
        val63 = ann3.pop("val6")
        val63c = ann3c.pop("val6")

        self.assertEqual(ann1, self.base1.annotations)
        self.assertNotEqual(ann1c, self.base1.annotations)
        self.assertEqual(ann2c, self.base2.annotations)
        self.assertEqual(ann3c, self.base3.annotations)
        self.assertEqual(targ, self.base1.annotations)

        assert_arrays_equal(val61, val6t)
        self.assertRaises(AssertionError, assert_arrays_equal, val61c, val6t)
        assert_arrays_equal(val62, val62c)
        assert_arrays_equal(val63, val63c)

        self.assertEqual(self.name1, self.base1.name)
        self.assertEqual(self.name2, self.base2.name)
        self.assertEqual(self.name3, self.base3.name)
        self.assertEqual(self.description1, self.base1.description)
        self.assertEqual(self.description2, self.base2.description)
        self.assertEqual(self.description3, self.base3.description)

    def test_merge_annotations__func__dict(self):
        ann1 = {
            "val0": "val0",
            "val1": 1,
            "val2": 2.2,
            "val3": "test1",
            "val4": [0.4],
            "val5": {0: 0, 1: {0: 0}},
            "val6": np.array([0, 1, 2]),
        }
        ann2 = {
            "val2": 2.2,
            "val3": "test2",
            "val4": [4, 4.4],
            "val5": {1: {1: 1}, 2: 2},
            "val6": np.array([4, 5, 6]),
            "val7": True,
        }

        ann1c = ann1.copy()
        ann2c = ann2.copy()

        targ = {
            "val0": "val0",
            "val1": 1,
            "val2": 2.2,
            "val3": "test1;test2",
            "val4": [0.4, 4, 4.4],
            "val5": {0: 0, 1: {0: 0, 1: 1}, 2: 2},
            "val7": True,
        }

        res = merge_annotations(ann1, ann2)

        val6t = np.array([0, 1, 2, 4, 5, 6])
        val6r = res.pop("val6")
        val61 = ann1.pop("val6")
        val61c = ann1c.pop("val6")
        val62 = ann2.pop("val6")
        val62c = ann2c.pop("val6")

        self.assertEqual(ann1, ann1c)
        self.assertEqual(ann2, ann2c)
        self.assertEqual(res, targ)

        assert_arrays_equal(val6r, val6t)
        self.assertRaises(AssertionError, assert_arrays_equal, val61, val6t)
        assert_arrays_equal(val61, val61c)
        assert_arrays_equal(val62, val62c)

    def test_merge_multiple_annotations__func__dict(self):
        ann1 = {
            "val0": "val0",
            "val1": 1,
            "val2": 2.2,
            "val3": "test1",
            "val4": [0.4],
            "val5": {0: 0, 1: {0: 0}},
            "val6": np.array([0, 1, 2]),
        }
        ann2 = {
            "val2": 2.2,
            "val3": "test2",
            "val4": [4, 4.4],
            "val5": {1: {1: 1}, 2: 2},
            "val6": np.array([4, 5, 6]),
            "val7": True,
        }
        ann3 = {
            "val2": 2.2,
            "val3": "test3",
            "val4": [44],
            "val5": {1: {2: 2}, 2: 2, 3: 3},
            "val6": np.array([8, 9, 10]),
            "val8": False,
        }

        ann1c = ann1.copy()
        ann2c = ann2.copy()
        ann3c = ann3.copy()

        targ = {
            "val0": "val0",
            "val1": 1,
            "val2": 2.2,
            "val3": "test1;test2;test3",
            "val4": [0.4, 4, 4.4, 44],
            "val5": {0: 0, 1: {0: 0, 1: 1, 2: 2}, 2: 2, 3: 3},
            "val7": True,
            "val8": False,
        }

        res = merge_annotations(ann1, ann2, ann3)

        val6t = np.array([0, 1, 2, 4, 5, 6, 8, 9, 10])
        val6r = res.pop("val6")
        val61 = ann1.pop("val6")
        val61c = ann1c.pop("val6")
        val62 = ann2.pop("val6")
        val62c = ann2c.pop("val6")
        val63 = ann3.pop("val6")
        val63c = ann3c.pop("val6")

        self.assertEqual(ann1, ann1c)
        self.assertEqual(ann2, ann2c)
        self.assertEqual(ann3, ann3c)
        self.assertEqual(res, targ)

        assert_arrays_equal(val6r, val6t)
        self.assertRaises(AssertionError, assert_arrays_equal, val61, val6t)
        assert_arrays_equal(val61, val61c)
        assert_arrays_equal(val62, val62c)
        assert_arrays_equal(val63, val63c)

    def test_merge_annotation__func__str(self):
        ann1 = "test1"
        ann2 = "test2"

        targ = "test1;test2"

        res = merge_annotation(ann1, ann2)

        self.assertEqual(res, targ)

    def test_merge_annotation__func__ndarray(self):
        ann1 = np.array([0, 1, 2])
        ann2 = np.array([4, 5, 6])

        ann1c = ann1.copy()
        ann2c = ann2.copy()

        targ = np.array([0, 1, 2, 4, 5, 6])

        res = merge_annotation(ann1, ann2)

        assert_arrays_equal(res, targ)
        assert_arrays_equal(ann1, ann1c)
        assert_arrays_equal(ann2, ann2c)

    def test_merge_annotation__func__list(self):
        ann1 = [0, 1, 2]
        ann2 = [4, 5, 6]

        ann1c = ann1[:]
        ann2c = ann2[:]

        targ = [0, 1, 2, 4, 5, 6]

        res = merge_annotation(ann1, ann2)

        self.assertEqual(res, targ)
        self.assertEqual(ann1, ann1c)
        self.assertEqual(ann2, ann2c)

    def test_merge_annotation__func__dict(self):
        ann1 = {0: 0, 1: {0: 0}}
        ann2 = {1: {1: 1}, 2: 2}

        ann1c = ann1.copy()
        ann2c = ann2.copy()

        targ = {0: 0, 1: {0: 0, 1: 1}, 2: 2}

        res = merge_annotation(ann1, ann2)

        self.assertEqual(res, targ)
        self.assertEqual(ann1, ann1c)
        self.assertEqual(ann2, ann2c)

    def test_merge_annotation__func__int(self):
        ann1 = 1
        ann2 = 1
        ann3 = 3

        targ = 1

        res = merge_annotation(ann1, ann2)

        self.assertEqual(res, targ)
        self.assertRaises(ValueError, merge_annotation, ann1, ann3)

    def test_merge_annotation__func__float(self):
        ann1 = 1.1
        ann2 = 1.1
        ann3 = 1.3

        targ = 1.1

        res = merge_annotation(ann1, ann2)

        self.assertEqual(res, targ)
        self.assertRaises(ValueError, merge_annotation, ann1, ann3)

    def test_merge_annotation__func__bool(self):
        ann1 = False
        ann2 = False
        ann3 = True
        ann4 = True

        targ1 = False
        targ2 = True

        res1 = merge_annotation(ann1, ann2)
        res2 = merge_annotation(ann3, ann4)

        self.assertEqual(res1, targ1)
        self.assertEqual(res2, targ2)
        self.assertRaises(ValueError, merge_annotation, ann1, ann3)
        self.assertRaises(ValueError, merge_annotation, ann2, ann4)

    def test_merge__dict(self):
        self.base1.annotations = {"val0": "val0", "val1": 1, "val2": 2.2, "val3": "test1"}
        self.base2.annotations = {"val2": 2.2, "val3": "test2", "val4": [4, 4.4], "val5": True}

        ann1 = self.base1.annotations
        ann1c = self.base1.annotations.copy()
        ann2c = self.base2.annotations.copy()

        targ = {"val0": "val0", "val1": 1, "val2": 2.2, "val3": "test1;test2", "val4": [4, 4.4], "val5": True}

        self.base1.merge(self.base2)

        self.assertEqual(ann1, self.base1.annotations)
        self.assertNotEqual(ann1c, self.base1.annotations)
        self.assertEqual(ann2c, self.base2.annotations)
        self.assertEqual(targ, self.base1.annotations)

        self.assertEqual(self.name1, self.base1.name)
        self.assertEqual(self.name2, self.base2.name)
        self.assertEqual(self.description1, self.base1.description)
        self.assertEqual(self.description2, self.base2.description)

    def test_merge_multiple__dict(self):
        self.base1.annotations = {"val0": "val0", "val1": 1, "val2": 2.2, "val3": "test1"}
        self.base2.annotations = {"val2": 2.2, "val3": "test2", "val4": [4, 4.4], "val5": True}
        self.base3.annotations = {"val2": 2.2, "val3": "test3", "val4": [44], "val5": True, "val6": False}

        ann1 = self.base1.annotations
        ann1c = self.base1.annotations.copy()
        ann2c = self.base2.annotations.copy()
        ann3c = self.base3.annotations.copy()

        targ = {
            "val0": "val0",
            "val1": 1,
            "val2": 2.2,
            "val3": "test1;test2;test3",
            "val4": [4, 4.4, 44],
            "val5": True,
            "val6": False,
        }

        self.base1.merge(self.base2, self.base3)

        self.assertEqual(ann1, self.base1.annotations)
        self.assertNotEqual(ann1c, self.base1.annotations)
        self.assertEqual(ann2c, self.base2.annotations)
        self.assertEqual(ann3c, self.base3.annotations)
        self.assertEqual(targ, self.base1.annotations)

        self.assertEqual(self.name1, self.base1.name)
        self.assertEqual(self.name2, self.base2.name)
        self.assertEqual(self.name3, self.base3.name)
        self.assertEqual(self.description1, self.base1.description)
        self.assertEqual(self.description2, self.base2.description)
        self.assertEqual(self.description3, self.base3.description)

    def test_merge_annotations__different_type_AssertionError(self):
        self.base1.annotations = {"val1": 1, "val2": 2.2, "val3": "tester"}
        self.base2.annotations = {"val3": False, "val4": [4, 4.4], "val5": True}
        self.base1.merge_annotations(self.base2)
        self.assertEqual(
            self.base1.annotations, {"val1": 1, "val2": 2.2, "val3": "MERGE CONFLICT", "val4": [4, 4.4], "val5": True}
        )

    def test_merge_multiple_annotations__different_type_AssertionError(self):
        self.base1.annotations = {"val1": 1, "val2": 2.2, "val3": "tester"}
        self.base2.annotations = {"val3": False, "val4": [4, 4.4], "val5": True}
        self.base3.annotations = {"val5": 1, "val6": 79, "val7": True}
        self.base1.merge_annotations(self.base2, self.base3)
        self.assertEqual(
            self.base1.annotations,
            {
                "val1": 1,
                "val2": 2.2,
                "val3": "MERGE CONFLICT",
                "val4": [4, 4.4],
                "val5": "MERGE CONFLICT",
                "val6": 79,
                "val7": True,
            },
        )

    def test_merge__different_type_AssertionError(self):
        self.base1.annotations = {"val1": 1, "val2": 2.2, "val3": "tester"}
        self.base2.annotations = {"val3": False, "val4": [4, 4.4], "val5": True}
        self.base1.merge(self.base2)
        self.assertEqual(
            self.base1.annotations, {"val1": 1, "val2": 2.2, "val3": "MERGE CONFLICT", "val4": [4, 4.4], "val5": True}
        )

    def test_merge_multiple__different_type_AssertionError(self):
        self.base1.annotations = {"val1": 1, "val2": 2.2, "val3": "tester"}
        self.base2.annotations = {"val3": False, "val4": [4, 4.4], "val5": True}
        self.base3.annotations = {"val5": 3.1, "val6": False, "val7": "val7"}
        self.base1.merge(self.base2, self.base3)
        self.assertEqual(
            self.base1.annotations,
            {
                "val1": 1,
                "val2": 2.2,
                "val3": "MERGE CONFLICT",
                "val4": [4, 4.4],
                "val5": "MERGE CONFLICT",
                "val6": False,
                "val7": "val7",
            },
        )

    def test_merge_annotations__unmergable_unequal_AssertionError(self):
        self.base1.annotations = {"val1": 1, "val2": 2.2, "val3": True}
        self.base2.annotations = {"val3": False, "val4": [4, 4.4], "val5": True}
        self.base1.merge_annotations(self.base2)
        self.assertEqual(
            self.base1.annotations, {"val1": 1, "val2": 2.2, "val3": "MERGE CONFLICT", "val4": [4, 4.4], "val5": True}
        )

    def test_merge_multiple_annotations__unmergable_unequal_AssertionError(self):
        self.base1.annotations = {"val1": 1, "val2": 2.2, "val3": True}
        self.base2.annotations = {"val3": False, "val4": [4, 4.4], "val5": 3.5}
        self.base3.annotations = {"val5": 3.4, "val6": [4, 4.4], "val7": True}
        self.base1.merge_annotations(self.base2, self.base3)
        self.assertEqual(
            self.base1.annotations,
            {
                "val1": 1,
                "val2": 2.2,
                "val3": "MERGE CONFLICT",
                "val4": [4, 4.4],
                "val5": "MERGE CONFLICT",
                "val6": [4, 4.4],
                "val7": True,
            },
        )

    def test_merge__unmergable_unequal_AssertionError(self):
        self.base1.annotations = {"val1": 1, "val2": 2.2, "val3": True}
        self.base2.annotations = {"val3": False, "val4": [4, 4.4], "val5": True}
        self.base1.merge(self.base2)
        self.assertEqual(
            self.base1.annotations, {"val1": 1, "val2": 2.2, "val3": "MERGE CONFLICT", "val4": [4, 4.4], "val5": True}
        )

    def test_merge_multiple__unmergable_unequal_AssertionError(self):
        self.base1.annotations = {"val1": 1, "val2": 2.2, "val3": True}
        self.base2.annotations = {"val3": False, "val4": [4, 4.4], "val5": True}
        self.base3.annotations = {"val5": 3.4, "val6": [4, 4.4], "val7": True}
        self.base1.merge(self.base2, self.base3)
        self.assertEqual(
            self.base1.annotations,
            {
                "val1": 1,
                "val2": 2.2,
                "val3": "MERGE CONFLICT",
                "val4": [4, 4.4],
                "val5": "MERGE CONFLICT",
                "val6": [4, 4.4],
                "val7": True,
            },
        )


class TestBaseNeoCoreTypes(unittest.TestCase):
    """
    TestCase to make sure annotations are properly checked for core built-in
    python data types
    """

    def setUp(self):
        """create the instance to be tested, called before every test"""
        self.base = BaseNeo()

    def test_python_nonetype(self):
        """test to make sure None type data is accepted"""
        value = None
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertEqual(value, self.base.annotations["data"])
        self.assertDictEqual(result, self.base.annotations)

    def test_python_int(self):
        """test to make sure int type data is accepted"""
        value = 10
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertEqual(value, self.base.annotations["data"])
        self.assertDictEqual(result, self.base.annotations)

    def test_python_float(self):
        """test to make sure float type data is accepted"""
        value = 9.2
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertEqual(value, self.base.annotations["data"])
        self.assertDictEqual(result, self.base.annotations)

    def test_python_complex(self):
        """test to make sure complex type data is accepted"""
        value = complex(23.17, 11.29)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertEqual(value, self.base.annotations["data"])
        self.assertDictEqual(result, self.base.annotations)

    def test_python_string(self):
        """test to make sure string type data is accepted"""
        value = "this is a test"
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertEqual(value, self.base.annotations["data"])
        self.assertDictEqual(result, self.base.annotations)

    def test_python_unicode(self):
        """test to make sure unicode type data is accepted"""
        value = "this is also a test"
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertEqual(value, self.base.annotations["data"])
        self.assertDictEqual(result, self.base.annotations)

    def test_python_bytes(self):
        """test to make sure bytes type data is accepted"""
        value = bytes("1,2,3,4,5", encoding="ascii")
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertEqual(value, self.base.annotations["data"])
        self.assertDictEqual(result, self.base.annotations)


class TestBaseNeoStandardLibraryTypes(unittest.TestCase):
    """
    TestCase to make sure annotations are properly checked for data types from
    the python standard library that are not core built-in data types
    """

    def setUp(self):
        """create the instance to be tested, called before every test"""
        self.base = BaseNeo()

    def test_python_fraction(self):
        """test to make sure Fraction type data is accepted"""
        value = Fraction(13, 21)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertEqual(value, self.base.annotations["data"])
        self.assertDictEqual(result, self.base.annotations)

    def test_python_decimal(self):
        """test to make sure Decimal type data is accepted"""
        value = Decimal("3.14")
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertEqual(value, self.base.annotations["data"])
        self.assertDictEqual(result, self.base.annotations)

    def test_python_datetime(self):
        """test to make sure datetime type data is accepted"""
        value = datetime(year=2008, month=12, day=3, hour=10, minute=4)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertEqual(value, self.base.annotations["data"])
        self.assertDictEqual(result, self.base.annotations)

    def test_python_date(self):
        """test to make sure date type data is accepted"""
        value = date(year=2008, month=12, day=3)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertEqual(value, self.base.annotations["data"])
        self.assertDictEqual(result, self.base.annotations)

    def test_python_time(self):
        """test to make sure time type data is accepted"""
        value = time(hour=10, minute=4)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertEqual(value, self.base.annotations["data"])
        self.assertDictEqual(result, self.base.annotations)

    def test_python_timedelta(self):
        """test to make sure timedelta type data is accepted"""
        value = timedelta(weeks=2, days=7, hours=18, minutes=28, seconds=18, milliseconds=28, microseconds=45)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertEqual(value, self.base.annotations["data"])
        self.assertDictEqual(result, self.base.annotations)


class TestBaseNeoContainerTypes(unittest.TestCase):
    """
    TestCase to make sure annotations are properly checked for data type
    inside python built-in container types
    """

    def setUp(self):
        """create the instance to be tested, called before every test"""
        self.base = BaseNeo()

    def test_python_list(self):
        """test to make sure list type data is accepted"""
        value = [
            None,
            10,
            9.2,
            complex(23, 11),
            ["this is a test", bytes("1,2,3,4,5", encoding="ascii")],
            [Fraction(13, 21), Decimal("3.14")],
        ]
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertListEqual(value, self.base.annotations["data"])
        self.assertDictEqual(result, self.base.annotations)

    def test_python_tuple(self):
        """test to make sure tuple type data is accepted"""
        value = (
            None,
            10,
            9.2,
            complex(23, 11),
            ("this is a test", bytes("1,2,3,4,5", encoding="ascii")),
            (Fraction(13, 21), Decimal("3.14")),
        )
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertTupleEqual(value, self.base.annotations["data"])
        self.assertDictEqual(result, self.base.annotations)

    def test_python_dict(self):
        """test to make sure dict type data is accepted"""
        value = {
            "NoneType": None,
            "int": 10,
            "float": 9.2,
            "complex": complex(23, 11),
            "dict1": {"string": "this is a test", "bytes": bytes("1,2,3,4,5", encoding="ascii")},
            "dict2": {"Fraction": Fraction(13, 21), "Decimal": Decimal("3.14")},
        }
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)

    def test_python_set(self):
        """test to make sure set type data is rejected"""
        value = {None, 10, 9.2, complex(23, 11)}
        self.assertRaises(ValueError, self.base.annotate, data=value)

    def test_python_frozenset(self):
        """test to make sure frozenset type data is rejected"""
        value = frozenset([None, 10, 9.2, complex(23, 11)])
        self.assertRaises(ValueError, self.base.annotate, data=value)

    def test_python_iter(self):
        """test to make sure iter type data is rejected"""
        value = iter([None, 10, 9.2, complex(23, 11)])
        self.assertRaises(ValueError, self.base.annotate, data=value)


class TestBaseNeoNumpyArrayTypes(unittest.TestCase):
    """
    TestCase to make sure annotations are properly checked for numpy arrays
    """

    def setUp(self):
        """create the instance to be tested, called before every test"""
        self.base = BaseNeo()

    def test_numpy_array_int(self):
        """test to make sure int type numpy arrays are accepted"""
        value = np.array([1, 2, 3, 4, 5], dtype=int)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)

    def test_numpy_array_uint(self):
        """test to make sure uint type numpy arrays are accepted"""
        value = np.array([1, 2, 3, 4, 5], dtype=np.uint)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)

    def test_numpy_array_intp(self):
        """test to make sure intp type numpy arrays are accepted"""
        value = np.array([1, 2, 3, 4, 5], dtype=np.intp)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)

    def test_numpy_array_uintp(self):
        """test to make sure uintp type numpy arrays are accepted"""
        value = np.array([1, 2, 3, 4, 5], dtype=np.uintp)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)

    def test_numpy_array_int8(self):
        """test to make sure int8 type numpy arrays are accepted"""
        value = np.array([1, 2, 3, 4, 5], dtype=np.int8)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)

    def test_numpy_array_uint8(self):
        """test to make sure uint8 type numpy arrays are accepted"""
        value = np.array([1, 2, 3, 4, 5], dtype=np.uint8)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)

    def test_numpy_array_int16(self):
        """test to make sure int16 type numpy arrays are accepted"""
        value = np.array([1, 2, 3, 4, 5], dtype=np.int16)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)

    def test_numpy_array_uint16(self):
        """test to make sure uint16 type numpy arrays are accepted"""
        value = np.array([1, 2, 3, 4, 5], dtype=np.uint16)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)

    def test_numpy_array_int32(self):
        """test to make sure int32 type numpy arrays are accepted"""
        value = np.array([1, 2, 3, 4, 5], dtype=np.int32)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)

    def test_numpy_array_uint32(self):
        """test to make sure uint32 type numpy arrays are accepted"""
        value = np.array([1, 2, 3, 4, 5], dtype=np.uint32)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)

    def test_numpy_array_int64(self):
        """test to make sure int64 type numpy arrays are accepted"""
        value = np.array([1, 2, 3, 4, 5], dtype=np.int64)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)

    def test_numpy_array_uint64(self):
        """test to make sure uint64 type numpy arrays are accepted"""
        value = np.array([1, 2, 3, 4, 5], dtype=np.uint64)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)

    def test_numpy_array_float(self):
        """test to make sure float type numpy arrays are accepted"""
        value = np.array([1, 2, 3, 4, 5], dtype=np.float64)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)

    def test_numpy_array_double(self):
        """test to make sure double type numpy arrays are accepted"""
        value = np.array([1, 2, 3, 4, 5], dtype=np.double)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)

    def test_numpy_array_float16(self):
        """test to make sure float16 type numpy arrays are accepted"""
        value = np.array([1, 2, 3, 4, 5], dtype=np.float16)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)

    def test_numpy_array_float32(self):
        """test to make sure float32 type numpy arrays are accepted"""
        value = np.array([1, 2, 3, 4, 5], dtype=np.float32)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)

    def test_numpy_array_float64(self):
        """test to make sure float64 type numpy arrays are accepted"""
        value = np.array([1, 2, 3, 4, 5], dtype=np.float64)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)

    @unittest.skipUnless(hasattr(np, "float128"), "float128 not available")
    def test_numpy_array_float128(self):
        """test to make sure float128 type numpy arrays are accepted"""
        value = np.array([1, 2, 3, 4, 5], dtype=np.float128)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)

    def test_numpy_array_complex(self):
        """test to make sure complex type numpy arrays are accepted"""
        value = np.array([1, 2, 3, 4, 5], complex)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)

    def test_numpy_scalar_complex64(self):
        """test to make sure complex64 type numpy arrays are accepted"""
        value = np.array([1, 2, 3, 4, 5], dtype=np.complex64)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)

    def test_numpy_scalar_complex128(self):
        """test to make sure complex128 type numpy arrays are accepted"""
        value = np.array([1, 2, 3, 4, 5], dtype=np.complex128)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)

    @unittest.skipUnless(hasattr(np, "complex256"), "complex256 not available")
    def test_numpy_scalar_complex256(self):
        """test to make sure complex256 type numpy arrays are accepted"""
        value = np.array([1, 2, 3, 4, 5], dtype=np.complex256)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)

    def test_numpy_array_bool(self):
        """test to make sure bool type numpy arrays are accepted"""
        value = np.array([1, 2, 3, 4, 5], dtype=bool)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)

    def test_numpy_array_str(self):
        """test to make sure str type numpy arrays are accepted"""
        value = np.array([1, 2, 3, 4, 5], dtype=str)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)

    def test_numpy_array_string0(self):
        """test to make sure string0 type numpy arrays are accepted"""
        value = np.array([1, 2, 3, 4, 5], dtype=np.str_)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)


class TestBaseNeoNumpyScalarTypes(unittest.TestCase):
    """
    TestCase to make sure annotations are properly checked for numpy scalars
    """

    def setUp(self):
        """create the instance to be tested, called before every test"""
        self.base = BaseNeo()

    def test_numpy_scalar_int(self):
        """test to make sure int type numpy scalars are accepted"""
        value = np.array(99, dtype=int)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)

    def test_numpy_scalar_uint(self):
        """test to make sure uint type numpy scalars are accepted"""
        value = np.array(99, dtype=np.uint)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)

    def test_numpy_scalar_intp(self):
        """test to make sure intp type numpy scalars are accepted"""
        value = np.array(99, dtype=np.intp)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)

    def test_numpy_scalar_uintp(self):
        """test to make sure uintp type numpy scalars are accepted"""
        value = np.array(99, dtype=np.uintp)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)

    def test_numpy_scalar_int8(self):
        """test to make sure int8 type numpy scalars are accepted"""
        value = np.array(99, dtype=np.int8)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)

    def test_numpy_scalar_uint8(self):
        """test to make sure uint8 type numpy scalars are accepted"""
        value = np.array(99, dtype=np.uint8)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)

    def test_numpy_scalar_int16(self):
        """test to make sure int16 type numpy scalars are accepted"""
        value = np.array(99, dtype=np.int16)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)

    def test_numpy_scalar_uint16(self):
        """test to make sure uint16 type numpy scalars are accepted"""
        value = np.array(99, dtype=np.uint16)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)

    def test_numpy_scalar_int32(self):
        """test to make sure int32 type numpy scalars are accepted"""
        value = np.array(99, dtype=np.int32)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)

    def test_numpy_scalar_uint32(self):
        """test to make sure uint32 type numpy scalars are accepted"""
        value = np.array(99, dtype=np.uint32)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)

    def test_numpy_scalar_int64(self):
        """test to make sure int64 type numpy scalars are accepted"""
        value = np.array(99, dtype=np.int64)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)

    def test_numpy_scalar_uint64(self):
        """test to make sure uint64 type numpy scalars are accepted"""
        value = np.array(99, dtype=np.uint64)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)

    def test_numpy_scalar_float(self):
        """test to make sure float type numpy scalars are accepted"""
        value = np.array(99, dtype=np.float64)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)

    def test_numpy_scalar_double(self):
        """test to make sure double type numpy scalars are accepted"""
        value = np.array(99, dtype=np.double)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)

    def test_numpy_scalar_float16(self):
        """test to make sure float16 type numpy scalars are accepted"""
        value = np.array(99, dtype=np.float16)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)

    def test_numpy_scalar_float32(self):
        """test to make sure float32 type numpy scalars are accepted"""
        value = np.array(99, dtype=np.float32)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)

    def test_numpy_scalar_float64(self):
        """test to make sure float64 type numpy scalars are accepted"""
        value = np.array(99, dtype=np.float64)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)

    @unittest.skipUnless(hasattr(np, "float128"), "float128 not available")
    def test_numpy_scalar_float128(self):
        """test to make sure float128 type numpy scalars are accepted"""
        value = np.array(99, dtype=np.float128)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)

    def test_numpy_scalar_complex(self):
        """test to make sure complex type numpy scalars are accepted"""
        value = np.array(99, dtype=complex)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)

    def test_numpy_scalar_complex64(self):
        """test to make sure complex64 type numpy scalars are accepted"""
        value = np.array(99, dtype=np.complex64)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)

    def test_numpy_scalar_complex128(self):
        """test to make sure complex128 type numpy scalars are accepted"""
        value = np.array(99, dtype=np.complex128)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)

    @unittest.skipUnless(hasattr(np, "complex256"), "complex256 not available")
    def test_numpy_scalar_complex256(self):
        """test to make sure complex256 type numpy scalars are accepted"""
        value = np.array(99, dtype=np.complex256)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)

    def test_numpy_scalar_bool(self):
        """test to make sure bool type numpy scalars are rejected"""
        value = np.array(99, dtype=bool)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)

    def test_numpy_array_str(self):
        """test to make sure str type numpy scalars are accepted"""
        value = np.array(99, dtype=str)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)

    def test_numpy_scalar_string0(self):
        """test to make sure string0 type numpy scalars are rejected"""
        value = np.array(99, dtype=np.str_)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)


class TestBaseNeoQuantitiesArrayTypes(unittest.TestCase):
    """
    TestCase to make sure annotations are properly checked for quantities
    arrays
    """

    def setUp(self):
        """create the instance to be tested, called before every test"""
        self.base = BaseNeo()

    def test_quantities_array_int(self):
        """test to make sure int type quantites arrays are accepted"""
        value = pq.Quantity([1, 2, 3, 4, 5], dtype=int, units=pq.s)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)

    def test_quantities_array_uint(self):
        """test to make sure uint type quantites arrays are accepted"""
        value = pq.Quantity([1, 2, 3, 4, 5], dtype=np.uint, units=pq.meter)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)

    def test_quantities_array_float(self):
        """test to make sure float type quantites arrays are accepted"""
        value = [1, 2, 3, 4, 5] * pq.kg
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)

    def test_quantities_array_str(self):
        """test to make sure str type quantites arrays are accepted"""
        value = pq.Quantity([1, 2, 3, 4, 5], dtype=str, units=pq.meter)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)


class TestBaseNeoQuantitiesScalarTypes(unittest.TestCase):
    """
    TestCase to make sure annotations are properly checked for quantities
    scalars
    """

    def setUp(self):
        """create the instance to be tested, called before every test"""
        self.base = BaseNeo()

    def test_quantities_scalar_int(self):
        """test to make sure int type quantites scalars are accepted"""
        value = pq.Quantity(99, dtype=int, units=pq.s)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)

    def test_quantities_scalar_uint(self):
        """test to make sure uint type quantites scalars are accepted"""
        value = pq.Quantity(99, dtype=np.uint, units=pq.meter)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)

    def test_quantities_scalar_float(self):
        """test to make sure float type quantites scalars are accepted"""
        value = 99 * pq.kg
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)

    def test_quantities_scalar_str(self):
        """test to make sure str type quantites scalars are accepted"""
        value = pq.Quantity(99, dtype=str, units=pq.meter)
        self.base.annotate(data=value)
        result = {"data": value}
        self.assertDictEqual(result, self.base.annotations)


class TestBaseNeoUserDefinedTypes(unittest.TestCase):
    """
    TestCase to make sure annotations are properly checked for arbitrary
    objects
    """

    def setUp(self):
        """create the instance to be tested, called before every test"""
        self.base = BaseNeo()

    def test_my_class(self):
        """test to make sure user defined class type data is rejected"""

        class Foo:
            pass

        value = Foo()
        self.assertRaises(ValueError, self.base.annotate, data=value)

    def test_my_class_list(self):
        """test to make sure user defined class type data is rejected"""

        class Foo:
            pass

        value = [Foo(), Foo(), Foo()]
        self.assertRaises(ValueError, self.base.annotate, data=value)


@unittest.skipUnless(HAVE_IPYTHON, "requires IPython")
class Test_pprint(unittest.TestCase):
    def test__pretty(self):
        name = "an object"
        description = "this is a test"
        obj = BaseNeo(name=name, description=description)
        res = pretty(obj)
        targ = f"BaseNeo name: '{name}' description: '{description}'"
        self.assertEqual(res, targ)


class Test_intersect_annotations(unittest.TestCase):
    """
    TestCase for intersect_annotations
    """

    def setUp(self):
        self.dict1 = {1: "1", 2: "2"}
        self.dict2 = {1: "1"}
        self.dict3 = {"list1": [1, 2, 3]}
        self.dict4 = {"list1": [1, 2, 3], "list2": [1, 2, 3]}
        self.dict5 = {"list1": [1, 2]}
        self.dict6 = {"array1": np.array([1, 2])}
        self.dict7 = {"array1": np.array([1, 2]), "array2": np.array([1, 2]), "array3": np.array([1, 2, 3])}

        self.all_simple_dicts = [
            self.dict1,
            self.dict2,
            self.dict3,
            self.dict4,
            self.dict5,
        ]

    def test_simple(self):
        result = intersect_annotations(self.dict1, self.dict2)
        self.assertDictEqual(self.dict2, result)

    def test_intersect_self(self):
        for d in self.all_simple_dicts:
            result = intersect_annotations(d, d)
            self.assertDictEqual(d, result)

    def test_list(self):
        result = intersect_annotations(self.dict3, self.dict4)
        self.assertDictEqual(self.dict3, result)

    def test_list_values(self):
        result = intersect_annotations(self.dict4, self.dict5)
        self.assertDictEqual({}, result)

    def test_keys(self):
        result = intersect_annotations(self.dict1, self.dict4)
        self.assertDictEqual({}, result)

    def test_arrays(self):
        result = intersect_annotations(self.dict6, self.dict7)
        self.assertEqual(self.dict6.keys(), result.keys())
        np.testing.assert_array_equal([1, 2], result["array1"])


if __name__ == "__main__":
    unittest.main()



"""
Tests of neo.io.neuroshareio
"""

import sys
import os
import pathlib
import shutil
import tarfile
import zipfile
import tempfile
import platform
import unittest
from urllib.request import urlretrieve

from neo.io import NeuroshareIO
from neo.test.iotest.common_io_test import BaseTestIO


class TestNeuroshareIO(BaseTestIO, unittest.TestCase):
    ioclass = NeuroshareIO
    entities_to_download = ["neuroshare"]
    entities_to_test = ["neuroshare/Multichannel_fil_1.mcd"]

    @classmethod
    def setUpClass(cls, *args, **kwargs):
        super().setUpClass()
        if sys.platform.startswith("win"):
            distantfile = (
                "http://download.multichannelsystems.com/download_data/software/neuroshare/nsMCDLibrary_3.7b.zip"
            )
            localfile = os.path.join(tempfile.gettempdir(), "nsMCDLibrary_3.7b.zip")
            if not os.path.exists(localfile):
                urlretrieve(distantfile, localfile)
            if platform.architecture()[0].startswith("64"):
                cls.dllname = os.path.join(
                    tempfile.gettempdir(), "Matlab/Matlab-Import-Filter/Matlab_Interface/nsMCDLibrary64.dll"
                )
                if not os.path.exists(cls.dllname):
                    zip = zipfile.ZipFile(localfile)
                    zip.extract(
                        "Matlab/Matlab-Import-Filter/Matlab_Interface/nsMCDLibrary64.dll", path=tempfile.gettempdir()
                    )
            else:
                cls.dllname = os.path.join(
                    tempfile.gettempdir(), "Matlab/Matlab-Import-Filter/Matlab_Interface/nsMCDLibrary.dll"
                )
                if not os.path.exists(cls.dllname):
                    zip = zipfile.ZipFile(localfile)
                    zip.extract(
                        "Matlab/Matlab-Import-Filter/Matlab_Interface/nsMCDLibrary.dll", path=tempfile.gettempdir()
                    )

        elif sys.platform.startswith("linux"):
            if platform.architecture()[0].startswith("64"):
                distantfile = "http://download.multichannelsystems.com/download_data/software/neuroshare/nsMCDLibrary_Linux64_3.7b.tar.gz"
                localfile = os.path.join(tempfile.gettempdir(), "nsMCDLibrary_Linux64_3.7b.tar.gz")
            else:
                distantfile = "http://download.multichannelsystems.com/download_data/software/neuroshare/nsMCDLibrary_Linux32_3.7b.tar.gz"
                localfile = os.path.join(tempfile.gettempdir(), "nsMCDLibrary_Linux32_3.7b.tar.gz")
            if not os.path.exists(localfile):
                urlretrieve(distantfile, localfile)
            cls.dllname = os.path.join(tempfile.gettempdir(), "nsMCDLibrary/nsMCDLibrary.so")
            if not os.path.exists(cls.dllname):
                tar = tarfile.open(localfile)
                tar.extract("nsMCDLibrary/nsMCDLibrary.so", path=tempfile.gettempdir())
        else:
            raise unittest.SkipTest("Not currently supported on OS X")

        # move dll to path discoverable by neuroshare package
        discoverable_path = pathlib.Path.home() / ".neuroshare" / cls.dllname.split("/")[-1]
        if not discoverable_path.exists():
            discoverable_path.parent.mkdir(exist_ok=True)
            shutil.copyfile(cls.dllname, discoverable_path)
            cls._cleanup_library = True
        else:
            cls._cleanup_library = False

        cls.default_keyword_arguments["dllname"] = cls.dllname

    @classmethod
    def tearDownClass(cls) -> None:
        if cls._cleanup_library:
            discoverable_path = pathlib.Path.home() / ".neuroshare" / cls.dllname.split("/")[-1]
            discoverable_path.unlink(missing_ok=True)

    def test_with_multichannel(self):
        for filename in self.files_to_test:
            filename = self.get_local_path(filename)
            reader = NeuroshareIO(filename, self.dllname)
            blocks = reader.read()
            n = len(blocks[0].segments[0].analogsignals)
            assert n > 0, f"Expect signals in file {filename}"


if __name__ == "__main__":
    unittest.main()



"""
Class for reading MED (Multiscale Electrophysiology Data) Format.
    
Uses the dhn-med-py python package, created by Dark Horse Neuro, Inc.

Authors: Dan Crepeau, Matt Stead
"""

import numpy as np

from .baserawio import (
    BaseRawIO,
    _signal_channel_dtype,
    _signal_stream_dtype,
    _spike_channel_dtype,
    _event_channel_dtype,
)


class MedRawIO(BaseRawIO):
    """
    Class for reading MED (Multiscale Electrophysiology Data) Format.

    Uses the dhn-med-py MED python package (version >= 1.0.0), created by
    Dark Horse Neuro, Inc. and medformat.org.

    Parameters
    ----------
    dirname: str | Path | None, default: None
        The folder containing the data files to load
    password: str | None, default: None
        The password for the Med session
    keep_original_times: bool, default: False
        If True UTC timestamps are used and returned as seconds referenced
        to midnight 1 Jan 1970
        If False timestamps are referenced to the beginning of the session
        with the beginning being 0

    Notes
    -----
    Currently reads the entire MED session.  Every discontinuity is considered
    to be a new segment.  Channels are grouped by sampling frequency, to
    create streams.  In MED all channels will line up time-wise, so streams
    will span the entire recording, and continuous sections of those streams
    are divided up into segments.


    """

    extensions = ["medd", "rdat", "ridx"]
    rawmode = "one-dir"

    def __init__(self, dirname=None, password=None, keep_original_times=False, **kargs):
        BaseRawIO.__init__(self, **kargs)

        import dhn_med_py
        from dhn_med_py import MedSession

        self.dirname = str(dirname)
        self.password = password
        self.keep_original_times = keep_original_times

    def _source_name(self):
        return self.dirname

    def _parse_header(self):

        import dhn_med_py
        from dhn_med_py import MedSession

        # Set a default password to improve compatibility and ease-of-use.
        # This password will be ignored if an unencrypted MED dataset is being used.
        if self.password is None:
            self.password = "L2_password"

        # Open the MED session (open data file pointers and read metadata files)
        self.sess = MedSession(self.dirname, self.password)

        # set the matrix calls to be "sample_major" as opposed to "channel_major"
        self.sess.set_major_dimension("sample")

        # find number of segments
        sess_contigua = self.sess.session_info["contigua"]
        self._nb_segment = len(sess_contigua)

        # find overall session start time.
        self._session_start_time = sess_contigua[0]["start_time"]

        # keep track of offset from metadata, if we are keeping original times.
        if not self.keep_original_times:
            self._session_time_offset = 0 - self._session_start_time
        else:
            self._session_time_offset = self.sess.session_info["metadata"]["recording_time_offset"]

        # find start/stop times of each segment
        self._seg_t_starts = []
        self._seg_t_stops = []
        for seg_idx in range(self._nb_segment):
            self._seg_t_starts.append(sess_contigua[seg_idx]["start_time"])
            self._seg_t_stops.append(sess_contigua[seg_idx]["end_time"])

        # find number of streams per segment
        self._stream_info = []
        self._num_stream_info = 0
        for chan_idx, chan_info in enumerate(self.sess.session_info["channels"]):
            chan_freq = chan_info["metadata"]["sampling_frequency"]

            # set MED session reference channel to be this channel, so the correct contigua is returned
            self.sess.set_reference_channel(chan_info["metadata"]["channel_name"])
            contigua = self.sess.find_discontinuities()

            # find total number of samples in this channel
            chan_num_samples = 0
            for seg_idx in range(len(contigua)):
                chan_num_samples += (contigua[seg_idx]["end_index"] - contigua[seg_idx]["start_index"]) + 1

            # see if we need a new stream, or add channel to existing stream
            add_to_existing_stream_info = False
            for stream_info in self._stream_info:
                if chan_freq == stream_info["sampling_frequency"] and chan_num_samples == stream_info["num_samples"]:
                    # found a match, so add it!
                    add_to_existing_stream_info = True
                    stream_info["chan_list"].append((chan_idx, chan_info["metadata"]["channel_name"]))
                    stream_info["raw_chans"].append(chan_info["metadata"]["channel_name"])
                    break

            if not add_to_existing_stream_info:
                self._num_stream_info += 1

                new_stream_info = {
                    "sampling_frequency": chan_info["metadata"]["sampling_frequency"],
                    "chan_list": [(chan_idx, chan_info["metadata"]["channel_name"])],
                    "contigua": contigua,
                    "raw_chans": [chan_info["metadata"]["channel_name"]],
                    "num_samples": chan_num_samples,
                }

                self._stream_info.append(new_stream_info)

        self.num_channels_in_session = len(self.sess.session_info["channels"])
        self.num_streams_in_session = self._num_stream_info

        signal_streams = []
        signal_channels = []

        # fill in signal_streams and signal_channels info
        for signal_stream_counter, stream_info in enumerate(self._stream_info):

            # get the stream start time, which is the start time of the first continuous section
            stream_start_time = (stream_info["contigua"][0]["start_time"] + self._session_time_offset) / 1e6

            # create stream name/id with info that we now have
            name = f'stream (rate,#sample,t0): ({stream_info["sampling_frequency"]}, {stream_info["num_samples"]}, {stream_start_time})'
            stream_id = signal_stream_counter
            signal_streams.append((name, stream_id))

            # add entry for signal_channels for each channel in a stream
            for chan in stream_info["chan_list"]:
                signal_channels.append(
                    (chan[1], chan[0], stream_info["sampling_frequency"], "int32", "uV", 1, 0, stream_id)
                )

        signal_streams = np.array(signal_streams, dtype=_signal_stream_dtype)
        signal_channels = np.array(signal_channels, dtype=_signal_channel_dtype)

        # no unit/epoch information contained in MED
        spike_channels = []
        spike_channels = np.array(spike_channels, dtype=_spike_channel_dtype)

        # Events
        event_channels = []
        event_channels.append(("Event", "event_channel", "event"))
        event_channels.append(("Epoch", "epoch_channel", "epoch"))
        event_channels = np.array(event_channels, dtype=_event_channel_dtype)

        # Create Neo header structure
        self.header = {}
        self.header["nb_block"] = 1
        self.header["nb_segment"] = [self._nb_segment]
        self.header["signal_streams"] = signal_streams
        self.header["signal_channels"] = signal_channels
        self.header["spike_channels"] = spike_channels
        self.header["event_channels"] = event_channels

        # `_generate_minimal_annotations()` must be called to generate the nested
        # dict of annotations/array_annotations
        self._generate_minimal_annotations()

        # Add custom annotations for neo objects
        bl_ann = self.raw_annotations["blocks"][0]
        bl_ann["name"] = "MED Data Block"
        # The following adds all of the MED session_info to the block annotations,
        # which includes features like patient name, recording location, etc.
        bl_ann.update(self.sess.session_info)

        # Give segments unique names
        for i in range(self._nb_segment):
            seg_ann = bl_ann["segments"][i]
            seg_ann["name"] = "Seg #" + str(i) + " Block #0"

        # this pprint lines really help for understand the nested (and complicated sometimes) dict
        # from pprint import pprint
        # pprint(self.raw_annotations)

    def _segment_t_start(self, block_index, seg_index):
        return (self._seg_t_starts[seg_index] + self._session_time_offset) / 1e6

    def _segment_t_stop(self, block_index, seg_index):
        return (self._seg_t_stops[seg_index] + self._session_time_offset) / 1e6

    def _get_signal_size(self, block_index, seg_index, stream_index):
        stream_segment_contigua = self._stream_info[stream_index]["contigua"]
        return (stream_segment_contigua[seg_index]["end_index"] - stream_segment_contigua[seg_index]["start_index"]) + 1

    def _get_signal_t_start(self, block_index, seg_index, stream_index):
        return (self._seg_t_starts[seg_index] + self._session_time_offset) / 1e6

    def _get_analogsignal_chunk(self, block_index, seg_index, i_start, i_stop, stream_index, channel_indexes):

        import dhn_med_py
        from dhn_med_py import MedSession

        # Correct for None start/stop inputs
        if i_start is None:
            i_start = 0
        if i_stop is None:
            i_stop = self.get_signal_size(block_index=block_index, seg_index=seg_index, stream_index=stream_index)

        # Check for invalid start/stop inputs
        if i_start < 0 or i_stop > self.get_signal_size(
            block_index=block_index, seg_index=seg_index, stream_index=stream_index
        ):
            raise IndexError("MED read error: Too many samples requested!")
        if i_start > i_stop:
            raise IndexError("MED read error: i_start (" + i_start + ") is greater than i_stop (" + i_stop + ")")

        num_channels = 0
        if channel_indexes is None:
            self.sess.set_channel_inactive("all")
            self.sess.set_channel_active(self._stream_info[stream_index]["raw_chans"])
            num_channels = len(self._stream_info[stream_index]["raw_chans"])
            self.sess.set_reference_channel(self._stream_info[stream_index]["raw_chans"][0])
        else:
            if any(channel_indexes < 0):
                raise IndexError(f"Can not index negative channels: {channel_indexes}")
            # Set all channels to be inactive, then selectively set some of them to be active
            self.sess.set_channel_inactive("all")
            for i, channel_idx in enumerate(channel_indexes):
                num_channels += 1
                self.sess.set_channel_active(self._stream_info[stream_index]["raw_chans"][channel_idx])
            self.sess.set_reference_channel(self._stream_info[stream_index]["raw_chans"][channel_indexes[0]])

        # Return empty dataset if start/stop samples are equal
        if i_start == i_stop:
            raw_signals = np.zeros((0, num_channels), dtype="int32")
            return raw_signals

        # Otherwise, return the matrix 2D array returned by the MED library
        start_sample_offset = self._stream_info[stream_index]["contigua"][seg_index]["start_index"]
        self.sess.read_by_index(i_start + start_sample_offset, i_stop + start_sample_offset)

        raw_signals = np.empty((i_stop - i_start, num_channels), dtype=np.int32)
        for i, chan in enumerate(self.sess.data["channels"]):
            raw_signals[:, i] = chan["data"]

        return raw_signals

    def _spike_count(self, block_index, seg_index, spike_channel_index):
        return None

    def _get_spike_timestamps(self, block_index, seg_index, spike_channel_index, t_start, t_stop):
        return None

    def _rescale_spike_timestamp(self, spike_timestamps, dtype):
        return None

    def _get_spike_raw_waveforms(self, block_index, seg_index, spike_channel_index, t_start, t_stop):
        return None

    def _event_count(self, block_index, seg_index, event_channel_index):

        # there are no epochs to consider for MED in this interface
        if self.header["event_channels"]["type"][event_channel_index] == b"epoch":
            return 0

        records = self.sess.get_session_records()

        # Find segment boundaries
        ts0 = (self.segment_t_start(block_index, seg_index) * 1e6) - self._session_time_offset
        ts1 = (self.segment_t_stop(block_index, seg_index) * 1e6) - self._session_time_offset

        # Only count Note and Neuralynx type records.
        count = 0
        for record in records:
            if (record["type_string"] == "Note" or record["type_string"] == "NlxP") and (
                record["start_time"] >= ts0 and record["start_time"] < ts1
            ):
                count += 1

        return count

    def _get_event_timestamps(self, block_index, seg_index, event_channel_index, t_start, t_stop):

        # There are no epochs to consider for MED in this interface,
        # so just bail out if that's what's being asked for
        if self.header["event_channels"]["type"][event_channel_index] == b"epoch":
            return np.array([]), np.array([]), np.array([], dtype="U")

        if t_start is not None:
            start_time = (t_start * 1e6) - self._session_time_offset
        else:
            start_time = (self.segment_t_start(block_index, seg_index) * 1e6) - self._session_time_offset

        if t_stop is not None:
            end_time = (t_stop * 1e6) - self._session_time_offset
        else:
            end_time = (self.segment_t_stop(block_index, seg_index) * 1e6) - self._session_time_offset

        # Ask MED session for a list of events that match time parameters
        records = self.sess.get_session_records(start_time, end_time)

        # create a subset of only Note and Neuralynx type records
        records_subset = []
        for record in records:
            if record["type_string"] == "Note" or record["type_string"] == "NlxP":
                records_subset.append(record)
        records = records_subset

        # if no records match our criteria, then we are done, output empty arrays
        if len(records) == 0:
            return np.array([]), np.array([]), np.array([], dtype="U")

        # inialize output arrays
        times = np.empty(shape=[len(records)])
        durations = None
        labels = []

        # populate output arrays of times and labels
        for i, record in enumerate(records):
            times[i] = (record["start_time"] + self._session_time_offset) / 1e6
            if record["type_string"] == "Note":
                labels.append(record["text"])
            elif record["type_string"] == "NlxP":
                labels.append("NlxP subport: " + str(record["subport"]) + " value: " + str(record["value"]))

        labels = np.asarray(labels, dtype="U")

        return times, durations, labels

    def _rescale_event_timestamp(self, event_timestamps, dtype, event_channel_index):
        return np.asarray(event_timestamps, dtype=dtype)

    def _rescale_epoch_duration(self, raw_duration, dtype, event_channel_index):
        return np.asarray(raw_duration, dtype=dtype)

    def __del__(self):
        try:
            # Important to make sure the session is closed, since the MED library only allows
            # one session to be open at a time.
            self.sess.close()
            del self.sess
        except Exception:
            pass

    def close(self):
        try:
            # Important to make sure the session is closed, since the MED library only allows
            # one session to be open at a time.
            self.sess.close()
        except Exception:
            pass



"""
IO for reading MED datasets using dhn-med-py library.

dhn-med-py
https://medformat.org
https://pypi.org/project/dhn-med-py/

MED Format Specifications: https://medformat.org

Author: Dan Crepeau, Matt Stead
"""

from neo.io.basefromrawio import BaseFromRaw
from neo.rawio.medrawio import MedRawIO


class MedIO(MedRawIO, BaseFromRaw):
    """
    IO for reading MED datasets.
    """

    name = "MED IO"
    description = "IO for reading MED datasets"

    _prefered_signal_group_mode = "group-by-same-units"
    mode = "dir"

    def __init__(self, dirname=None, password=None, keep_original_times=False):
        MedRawIO.__init__(self, dirname=dirname, password=password, keep_original_times=keep_original_times)
        """
        Initialise IO instance

        Parameters
        ----------
        dirname : str
            Directory containing data files
        password : str
            MED sessions can be optionally encrypted with a password.
            Default: None
        keep_original_times : bool
            Preserve original time stamps as in data files. By default datasets are
            shifted to begin at t_start = 0.  When set to True, timestamps will be
            returned as UTC (seconds since midnight 1 Jan 1970).
            Default: False
        """
        BaseFromRaw.__init__(self, dirname)

    def close(self):
        MedRawIO.close(self)

    def __del__(self):
        MedRawIO.__del__(self)



import struct
import os
import math

import numpy as np

from .baseio import BaseIO
from neo.core import ImageSequence, Segment, Block


class BlkIO(BaseIO):
    """
    Neo IO module for optical imaging data stored as BLK file

    *Usage*:
        >>> from neo import io
        >>> import quantities as pq
        >>> r = io.BlkIO("file_blk_1.BLK",units='V',sampling_rate=1.0*pq.Hz,
        ...              spatial_scale=1.0*pq.Hz)
        >>> block = r.read_block()
        reading the header
        reading block
        returning block
        >>> block
        Block with 6 segments
        file_origin: 'file_blk_1.BLK'
        # segments (N=6)
        0: Segment with 1 imagesequences description: 'stim nb:0' # analogsignals (N=0)
        1: Segment with 1 imagesequences description: 'stim nb:1' # analogsignals (N=0)
        2: Segment with 1 imagesequences description: 'stim nb:2' # analogsignals (N=0)
        3: Segment with 1 imagesequences description: 'stim nb:3' # analogsignals (N=0)
        4: Segment with 1 imagesequences description: 'stim nb:4' # analogsignals (N=0)
        5: Segment with 1 imagesequences description: 'stim nb:5' # analogsignals (N=0)

    Many thanks to Thomas Deneux for the MATLAB code on which this was based.
    """

    name = "BLK IO"
    description = "Neo IO module for optical imaging data stored as BLK file"

    _prefered_signal_goup_mode = "group-by-same-units"
    is_readable = True
    is_wirtable = False

    supported_objects = [Block, Segment, ImageSequence]
    readble_objects = supported_objects

    support_lazy = False

    read_params = {}
    write_params = {}

    extensions = []

    mode = "file"

    def __init__(self, file_name=None, units=None, sampling_rate=None, spatial_scale=None, **kwargs):
        BaseIO.__init__(self, file_name, **kwargs)
        self.units = units
        self.sampling_rate = sampling_rate
        self.spatial_scale = spatial_scale

    def read(self, lazy=False, **kwargs):
        """
        Return all data from the file as a list of Blocks
        """
        if lazy:
            raise ValueError("This IO module does not support lazy loading")
        return [
            self.read_block(
                lazy=lazy,
                units=self.units,
                sampling_rate=self.sampling_rate,
                spatial_scale=self.spatial_scale,
                **kwargs,
            )
        ]

    def read_block(self, lazy=False, **kargs):

        def read(name, type, nb, dictionary, file):

            if type == "int32":
                # dictionary[name] = int.from_bytes(file.read(4), byteorder=sys.byteorder, signed=True)
                dictionary[name] = struct.unpack("i", file.read(4))[0]
            if type == "float32":
                dictionary[name] = struct.unpack("f", file.read(4))[0]
            if type == "uint8":
                l = []
                for i in range(nb):
                    l.append(chr(struct.unpack("B", file.read(1))[0]))
                dictionary[name] = l
            if type == "uint16":
                l = []
                for i in range(nb):
                    l.append((struct.unpack("H", file.read(2)))[0])
                dictionary[name] = l
            if type == "short":
                dictionary[name] = struct.unpack("h", file.read(2))[0]

            return dictionary

        def read_header(file_name):

            file = open(file_name, "rb")

            i = [
                ["file_size", "int32", 1],
                ["checksum_header", "int32", 1],
                ["check_data", "int32", 1],
                ["lenheader", "int32", 1],
                ["versionid", "float32", 1],
                ["filetype", "int32", 1],
                ["filesubtype", "int32", 1],
                ["datatype", "int32", 1],
                ["sizeof", "int32", 1],
                ["framewidth", "int32", 1],
                ["frameheight", "int32", 1],
                ["nframesperstim", "int32", 1],
                ["nstimuli", "int32", 1],
                ["initialxbinfactor", "int32", 1],
                ["initialybinfactor", "int32", 1],
                ["xbinfactor", "int32", 1],
                ["ybinfactor", "int32", 1],
                ["username", "uint8", 32],
                ["recordingdate", "uint8", 16],
                ["x1roi", "int32", 1],
                ["y1roi", "int32", 1],
                ["x2roi", "int32", 1],
                ["y2roi", "int32", 1],
                ["stimoffs", "int32", 1],
                ["stimsize", "int32", 1],
                ["frameoffs", "int32", 1],
                ["framesize", "int32", 1],
                ["refoffs", "int32", 1],
                ["refsize", "int32", 1],
                ["refwidth", "int32", 1],
                ["refheight", "int32", 1],
                ["whichblocks", "uint16", 16],
                ["whichframe", "uint16", 16],
                ["loclip", "int32", 1],
                ["hiclip", "int32", 1],
                ["lopass", "int32", 1],
                ["hipass", "int32", 1],
                ["operationsperformed", "uint8", 64],
                ["magnification", "float32", 1],
                ["gain", "uint16", 1],
                ["wavelength", "uint16", 1],
                ["exposuretime", "int32", 1],
                ["nrepetitions", "int32", 1],
                ["acquisitiondelay", "int32", 1],
                ["interstiminterval", "int32", 1],
                ["creationdate", "uint8", 16],
                ["datafilename", "uint8", 64],
                ["orareserved", "uint8", 256],
            ]

            dic = {}
            for x in i:
                dic = read(name=x[0], type=x[1], nb=x[2], dictionary=dic, file=file)

            if dic["filesubtype"] == 13:
                i = [
                    ["includesrefframe", "int32", 1],
                    ["temp", "uint8", 128],
                    ["ntrials", "int32", 1],
                    ["scalefactors", "int32", 1],
                    ["cameragain", "short", 1],
                    ["ampgain", "short", 1],
                    ["samplingrate", "short", 1],
                    ["average", "short", 1],
                    ["exposuretime", "short", 1],
                    ["samplingaverage", "short", 1],
                    ["presentaverage", "short", 1],
                    ["framesperstim", "short", 1],
                    ["trialsperblock", "short", 1],
                    ["sizeofanalogbufferinframes", "short", 1],
                    ["cameratrials", "short", 1],
                    ["filler", "uint8", 106],
                    ["dyedaqreserved", "uint8", 106],
                ]
                for x in i:
                    dic = read(name=x[0], type=x[1], nb=x[2], dictionary=dic, file=file)
                # nottested
                #  p.listofstimuli=temp(1:max(find(temp~=0)))';  % up to first non-zero stimulus
                dic["listofstimuli"] = dic["temp"][0 : np.argwhere(x != 0).max(0)]
            else:
                i = [
                    ["includesrefframe", "int32", 1],
                    ["listofstimuli", "uint8", 256],
                    ["nvideoframesperdataframe", "int32", 1],
                    ["ntrials", "int32", 1],
                    ["scalefactor", "int32", 1],
                    ["meanampgain", "float32", 1],
                    ["meanampdc", "float32", 1],
                    ["vdaqreserved", "uint8", 256],
                ]
                for x in i:
                    dic = read(name=x[0], type=x[1], nb=x[2], dictionary=dic, file=file)
            i = [["user", "uint8", 256], ["comment", "uint8", 256], ["refscalefactor", "int32", 1]]
            for x in i:
                dic = read(name=x[0], type=x[1], nb=x[2], dictionary=dic, file=file)
            dic["actuallength"] = os.stat(file_name).st_size
            file.close()

            return dic

        # start of the reading process
        nblocks = 1
        print("reading the header")
        header = read_header(self.filename)
        nstim = header["nstimuli"]
        ni = header["framewidth"]
        nj = header["frameheight"]
        nfr = header["nframesperstim"]
        lenh = header["lenheader"]
        framesize = header["framesize"]
        filesize = header["file_size"]
        dtype = header["datatype"]
        gain = header["meanampgain"]
        dc = header["meanampdc"]
        scalefactor = header["scalefactor"]

        # [["dtype","nbytes","datatype","type_out"],[...]]
        l = [
            [11, 1, "uchar", "uint8", "B"],
            [12, 2, "ushort", "uint16", "H"],
            [13, 4, "ulong", "uint32", "I"],
            [14, 4, "float", "single", "f"],
        ]

        for i in l:
            if dtype == i[0]:
                nbytes, datatype, type_out, struct_type = i[1], i[2], i[3], i[4]

        if framesize != ni * nj * nbytes:
            print("BAD HEADER!!! framesize does not match framewidth*frameheight*nbytes!")
            framesize = ni * nj * nbytes
        if (filesize - lenh) > (framesize * nfr * nstim):
            nfr2 = nfr + 1
            includesrefframe = True
        else:
            nfr2 = nfr
            includesrefframe = False

        nbin = nblocks
        conds = [i for i in range(1, nstim + 1)]
        ncond = len(conds)
        data = [[[np.zeros((ni, nj, nfr), type_out)] for x in range(ncond)] for i in range(nbin)]
        for k in range(1, nbin + 1):
            print("reading block")
            bin = np.arange(math.floor((k - 1 / nbin * nblocks) + 1), math.floor((k / nbin * nblocks) + 1))
            sbin = bin.size
            for j in range(1, sbin + 1):
                file = open(self.filename, "rb")
                for i in range(1, ncond + 1):

                    framestart = conds[i - 1] * nfr2 - nfr
                    offset = framestart * ni * nj * nbytes + lenh
                    file.seek(offset, 0)

                    a = [(struct.unpack(struct_type, file.read(nbytes)))[0] for m in range(ni * nj * nfr)]
                    a = np.reshape(np.array(a, dtype=type_out, order="F"), (ni * nj, nfr), order="F")
                    a = np.reshape(a, (ni, nj, nfr), order="F")

                    if includesrefframe:
                        # not tested
                        framestart = (conds[i] - 1) * nfr2
                        offset = framestart * ni * nj * nbytes + lenh

                        file.seek(offset)

                        ref = [(struct.unpack(struct_type, file.read(nbytes)))[0] for m in range(ni * nj)]
                        ref = np.array(ref, dtype=type_out)
                        for y in range(len(ref)):
                            ref[y] *= scalefactor
                        ref = np.reshape(ref, (ni, nj))
                        b = np.tile(ref, [1, 1, nfr])
                        for y in range(len(a)):
                            b.append([])
                            for x in range(len(a[y])):
                                b[y + 1].append([])
                                for frame in range(len(a[y][x])):
                                    b[y + 1][x][frame] = (a[y][x][frame] / gain) - (scalefactor * dc / gain)
                        a = b
                    if sbin == 1:
                        data[k - 1][i - 1] = a
                    else:
                        # not tested
                        for y in range(len(a)):
                            for x in range(len(a[y])):
                                a[y][x] /= sbin
                        data[k - 1][i - 1] = data[k - 1][i - 1] + a / sbin

                file.close()

        # data format [block][stim][width][height][frame]]
        # data structure should be [block][stim][frame][width][height] in order to be easy to use with neo
        # each file is a block
        # each stim could be a segment
        # then an image sequence [frame][width][height]
        # image need to be rotated

        # changing order of data for compatibility
        # [block][stim][width][height][frame]]
        # to
        # [block][stim][frame][width][height]

        for block in range(len(data)):
            for stim in range(len(data[block])):
                a = []
                for frame in range(header["nframesperstim"]):
                    a.append([])
                    for width in range(len(data[block][stim])):
                        a[frame].append([])
                        for height in range(len(data[block][stim][width])):
                            a[frame][width].append(data[block][stim][width][height][frame])
                    # rotation of data to be the same as thomas deneux screenshot
                    a[frame] = np.rot90(np.fliplr(a[frame]))
                data[block][stim] = a

        block = Block(file_origin=self.filename)
        for stim in range(len(data[0])):
            image_sequence = ImageSequence(
                data[0][stim], units=self.units, sampling_rate=self.sampling_rate, spatial_scale=self.spatial_scale
            )
            segment = Segment(file_origin=self.filename, description=("stim nb:" + str(stim)))
            segment.imagesequences = [image_sequence]
            for key in header:
                block.annotations[key] = header[key]
            block.segments.append(segment)

        print("returning block")

        return block



"""
Neo IO module for optical imaging data stored as a folder of TIFF images.
"""

import glob
import re

import numpy as np

from neo.core import ImageSequence, Segment, Block
from .baseio import BaseIO


class TiffIO(BaseIO):
    """
    Neo IO module for optical imaging data stored as a folder of TIFF images.

    *Usage*:
        >>> from neo import io
        >>> import quantities as pq
        >>> r = io.TiffIO("dir_tiff",spatial_scale=1.0*pq.mm, units='V',
        ...               sampling_rate=1.0*pq.Hz)
        >>> block = r.read_block()
        read block
        creating segment
        returning block
        >>> block
        Block with 1 segments
        file_origin: 'test'
        # segments (N=1)
        0: Segment with 1 imagesequences
            annotations: {'tiff_file_names': ['file_tif_1_.tiff',
                'file_tif_2.tiff',
                'file_tif_3.tiff',
                'file_tif_4.tiff',
                'file_tif_5.tiff',
                'file_tif_6.tiff',
                'file_tif_7.tiff',
                'file_tif_8.tiff',
                'file_tif_9.tiff',
                'file_tif_10.tiff',
                'file_tif_11.tiff',
                'file_tif_12.tiff',
                'file_tif_13.tiff',
                'file_tif_14.tiff']}
            # analogsignals (N=0)
    """

    name = "TIFF IO"
    description = "Neo IO module for optical imaging data stored as a folder of TIFF images."

    _prefered_signal_group_mode = "group-by-same-units"
    is_readable = True
    is_writable = False

    supported_objects = [Block, Segment, ImageSequence]
    readable_objects = supported_objects
    writeable_objects = []

    support_lazy = False

    read_params = {}
    write_params = {}

    extensions = []

    mode = "dir"

    def __init__(self, directory_path=None, units=None, sampling_rate=None, spatial_scale=None, **kwargs):
        import PIL

        BaseIO.__init__(self, directory_path, **kwargs)
        self.units = units
        self.sampling_rate = sampling_rate
        self.spatial_scale = spatial_scale

    def read_block(self, lazy=False, **kwargs):
        import PIL

        # to sort file
        def natural_sort(l):
            convert = lambda text: int(text) if text.isdigit() else text.lower()
            alphanum_key = lambda key: [convert(c) for c in re.split("([0-9]+)", key)]
            return sorted(l, key=alphanum_key)

        # find all the images in the given directory
        file_name_list = []
        # name of extensions to track
        types = ["*.tif", "*.tiff"]
        for file in types:
            file_name_list.append(glob.glob(self.filename + "/" + file))
        # flatten list
        file_name_list = [item for sublist in file_name_list for item in sublist]
        # delete path in the name of file
        file_name_list = [file_name[len(self.filename) + 1 : :] for file_name in file_name_list]
        # sorting file
        file_name_list = natural_sort(file_name_list)
        list_data_image = []
        for file_name in file_name_list:
            data = np.array(PIL.Image.open(self.filename + "/" + file_name)).astype(np.float32)
            list_data_image.append(data)
        list_data_image = np.array(list_data_image)
        if len(list_data_image.shape) == 4:
            list_data_image = []
            for file_name in file_name_list:
                image = PIL.Image.open(self.filename + "/" + file_name).convert("L")
                data = np.array(image).astype(np.float32)
                list_data_image.append(data)

        print("read block")
        image_sequence = ImageSequence(
            np.stack(list_data_image),
            units=self.units,
            sampling_rate=self.sampling_rate,
            spatial_scale=self.spatial_scale,
        )
        print("creating segment")
        segment = Segment(file_origin=self.filename)
        segment.annotate(tiff_file_names=file_name_list)
        segment.imagesequences = [image_sequence]

        block = Block(file_origin=self.filename)
        block.segments.append(segment)
        print("returning block")
        return block



import unittest

from neo.rawio.cedrawio import CedRawIO

from neo.test.rawiotest.common_rawio_test import BaseTestRawIO

try:
    import sonpy

    HAVE_SONPY = True
except ImportError:
    HAVE_SONPY = False


# This runs standard tests, this is mandatory for all IOs
@unittest.skipUnless(HAVE_SONPY, "requires sonpy package and all its dependencies")
class TestCedRawIO(
    BaseTestRawIO,
    unittest.TestCase,
):
    rawioclass = CedRawIO
    entities_to_test = ["spike2/m365_1sec.smrx", "spike2/File_spike2_1.smr", "spike2/Two-mice-bigfile-test000.smr"]
    entities_to_download = ["spike2"]


if __name__ == "__main__":
    unittest.main()



import unittest

from neo.rawio.elanrawio import ElanRawIO

from neo.test.rawiotest.common_rawio_test import BaseTestRawIO


class TestElanRawIO(
    BaseTestRawIO,
    unittest.TestCase,
):
    rawioclass = ElanRawIO
    entities_to_test = ["elan/File_elan_1.eeg"]
    entities_to_download = [
        "elan",
    ]


if __name__ == "__main__":
    unittest.main()



"""
Analyzing and Plotting Data with Neo Structures
===============================================
"""

######################################################
# First we import some packages. Since we are making simulated
# data we will import quite a few neo features as well as use
# quantities to provide units

from itertools import cycle
import numpy as np
from quantities import ms, mV, kHz
import matplotlib.pyplot as plt
from neo import Block, Segment, ChannelView, Group, SpikeTrain, AnalogSignal

##########################################################################
# For Neo we start with a block of data that will contain segments of data
# so we will create a block of probe data that has a couple tetrodes
# Then we will load in 3 segments (for examples trials of a stimulus)
store_signals = False

block = Block(name="probe data", tetrode_ids=["Tetrode #1", "Tetrode #2"])
block.segments = [
    Segment(name="trial #1", index=0),
    Segment(name="trial #2", index=1),
    Segment(name="trial #3", index=2),
]

# we will decide how many units each tetrode has found. If only science was this easy
n_units = {"Tetrode #1": 2, "Tetrode #2": 5}

##################################################################################
# Neo can also have groups. Groups are structures within a block that can cross segments
# for example we could group a neuron across trials or across probes.

# Create a group for each neuron, annotate each group with the tetrode from which it was recorded
groups = []
counter = 0
for tetrode_id, n in n_units.items():
    groups.extend([Group(name=f"neuron #{counter + i + 1}", tetrode_id=tetrode_id) for i in range(n)])
    counter += n
block.groups.extend(groups)

iter_group = cycle(groups)

##########################################################################################
# Segments are also containers of data. Segments can hold raw signal data like an AnalogSignal
# Segments can also hold spiketrain data (in a SpikeTrain). It can also hold event data (which
# we are not show in this example)


# Create dummy data, one segment at a time
for segment in block.segments:

    # create two 4-channel AnalogSignals with simulated data (because we have two tetrodes!)
    # note that the AnalogSignal with have numpy array-like data with units and sampling rates
    # Neo keeps track of these units while also giving you the flexibility of treating the raw data
    # like a numpy array
    signals = {
        "Tetrode #1": AnalogSignal(np.random.rand(1000, 4) * mV, sampling_rate=10 * kHz, tetrode_id="Tetrode #1"),
        "Tetrode #2": AnalogSignal(np.random.rand(1000, 4) * mV, sampling_rate=10 * kHz, tetrode_id="Tetrode #2"),
    }
    if store_signals:
        segment.analogsignals.extend(signals.values())

    # create spike trains with simulated data
    # we will pretend the spikes have been extracted from the simulated signal
    for tetrode_id in ("Tetrode #1", "Tetrode #2"):
        for i in range(n_units[tetrode_id]):
            spiketrain = SpikeTrain(np.random.uniform(0, 100, size=30) * ms, t_stop=100 * ms)
            # assign each spiketrain to the appropriate segment
            segment.spiketrains.append(spiketrain)
            # assign each spiketrain to a given neuron
            current_group = next(iter_group)
            current_group.add(spiketrain)
            if store_signals:
                # add to the group a reference to the signal from which the spikes were obtained
                # this does not give a 1:1 correspondance between spike trains and signals,
                # for that we could use additional groups (and have groups of groups)
                current_group.add(signals[tetrode_id])


###################################################
# Now we will plot the data
# Neo doesn't provide it's own plotting functions, but
# since its data can be treated like numpy arrays
# it is easy to use standard packages like matplotlib
# for all your plotting needs
# We do a classic in neuroscience and show various ways
# to plot a PSTH (Peristimulus histogram)

###################################################
# .. by trial
plt.figure()
for seg in block.segments:
    print(f"Analyzing segment {seg.index}")
    stlist = [st - st.t_start for st in seg.spiketrains]
    plt.subplot(len(block.segments), 1, seg.index + 1)
    count, bins = np.histogram(stlist)
    plt.bar(bins[:-1], count, width=bins[1] - bins[0])
    plt.title(f"PSTH in segment {seg.index}")
plt.tight_layout()
plt.show()

####################################################
# ..by neuron

plt.figure()
for i, group in enumerate(block.groups):
    stlist = [st - st.t_start for st in group.spiketrains]
    plt.subplot(len(block.groups), 1, i + 1)
    count, bins = np.histogram(stlist)
    plt.bar(bins[:-1], count, width=bins[1] - bins[0])
    plt.title(f"PSTH of unit {group.name}")
plt.tight_layout()
plt.show()

###########################################################
# ..by tetrode (or other electrode number)

plt.figure()
for i, tetrode_id in enumerate(block.annotations["tetrode_ids"]):
    stlist = []
    for unit in block.filter(objects=Group, tetrode_id=tetrode_id):
        stlist.extend([st - st.t_start for st in unit.spiketrains])
    plt.subplot(2, 1, i + 1)
    count, bins = np.histogram(stlist)
    plt.bar(bins[:-1], count, width=bins[1] - bins[0])
    plt.title(f"PSTH blend of tetrode {tetrode_id}")
plt.tight_layout()
plt.show()
